# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

# import hsr_spectrum.spectrumUtils as SpectrumUtils
from hsr1.utils.spectrum import spectrumUtils as SpectrumUtils

class PixelSpectrum:
    def __init__(self, cal_poly=None, poly_list=None):
        ##### calibration: contains the polynomial used for wavelength calibration, read from file
        self.cal_poly = cal_poly
        self.poly_list = poly_list
    
    def calculate_from_wavelengths(self, wavelength_spectra:[pd.DataFrame], data_cols):
        spectra = wavelength_spectra.copy()
        old_spectra = wavelength_spectra.copy()
        for i, channel in enumerate(data_cols):
            poly = self.cal_poly
            if self.poly_list is not None:
                poly = self.poly_list[i]
            sample_pixels = np.arange(0, 1200)
            sample_wavelengths = np.array([SpectrumUtils.apply_polynomial(pixel, poly[0], poly[1], poly[2], poly[3]) for pixel in sample_pixels])
            
            for j in range(len(spectra)):
                indecies = np.arange(300, 1101)
                spectra.loc[j, channel] = self.apply_wavelength_calibration(spectra.loc[j, channel], poly, indecies, flip=True)
                
        spectra["pc_time_end_measurement"] = old_spectra["pc_time_end_measurement"]
        return spectra



    def apply_wavelength_calibration(self, spectrum, poly, indecies=np.arange(1200), inverse=False, flip=None, output_shape=None, isdf=None, plot=False): 
        """takes a polynomial and shifts the spectrum according to that polynomial
        params:
            spectrum: the spectrum being converted
            poly: the polynomial calibration
            indecies: the size of the input array
            inverse: wether the polynomial is applied or is inverse, wavelength->pixel = True
            flip: some usecases require both inputs to be flipped so they are ascending, 
                    usually flip=True when applying inverse, and vice versa but there are some situations where this isnt the case
            output shape: [start, stop]the cropping to apply once calibrated
        
        to apply a pixel -> wavelength calibration, inverse=True and default values should work
        to apply a wavelength -> pixel calibration, use the constructor and calculate_from_wavelength
        to apply a wavelength -> wavelength calibration in index space, indecies should be np.arange(801) and output shape should be [0, 801]
        to apply a wavelength -> wavelength calibration in wavelength space, indecies should be np.arange(300, 1001) and output shape should be [300, 1101]
        
        """
        if output_shape is None:
            if inverse:
                output_shape = (300, 1101)
            else:
                output_shape = (0, 1200)
        
        if isdf is None:
            if isinstance(spectrum, pd.DataFrame):
                isdf = True
            else:
                isdf = False
        
        if isdf:
            ##### df of spectra
            spectrum = spectrum.copy()
            
            for i in range(len(spectrum)):
                one_spectrum = spectrum["global_spectrum"].iloc[i]
                if inverse:
                    if flip is None:
                        flip=True
                    corresponding_pixels = SpectrumUtils.apply_inverse_polynomial(indecies, poly[0], poly[1], poly[2], poly[3])
                else:
                    if flip is None:
                        flip=False
                    corresponding_pixels = SpectrumUtils.apply_polynomial(indecies, poly[0], poly[1], poly[2], poly[3])
                
                if flip:
                    corresponding_pixels = np.flip(corresponding_pixels)
                    one_spectrum = np.flip(one_spectrum)
                
                result = np.interp(np.arange(-1200, 2400), corresponding_pixels, one_spectrum)    
                output_shape = [output_shape[0]+1200, output_shape[1]+1200]
                spectrum.loc[i, "global_spectrum"] = result[output_shape[0]:output_shape[1]]
            return spectrum
        
        else:
            ##### single spectrum
            if inverse:
                if flip is None:
                    flip=True
                corresponding_pixels = SpectrumUtils.apply_inverse_polynomial(indecies, poly[0], poly[1], poly[2], poly[3])
            else:
                if flip is None:
                    flip=False
                corresponding_pixels = SpectrumUtils.apply_polynomial(indecies, poly[0], poly[1], poly[2], poly[3])
            
            if flip:
                corresponding_pixels = np.flip(corresponding_pixels)
                spectrum = np.flip(spectrum)
            
            
            result = np.interp(np.arange(-1200, 2400), corresponding_pixels, spectrum)
            
            if plot:
                plt.plot(indecies)
                plt.show()
                plt.plot(corresponding_pixels)
                plt.show()
                plt.plot(spectrum)
                plt.plot(result)
                plt.show()
            
            output_shape = [output_shape[0]+1200, output_shape[1]+1200]
            return result[output_shape[0]:output_shape[1]]
    
    
    def apply_constant_wavelength_calibration(self, spectrum, cal):
        """applied a constant shift left or right
        params:
            spectrum: the spectrum to be shifted
            cal: the amount to shift by
        returns:
            spectrum: shifted spectrum
        
        pads with zeros
        """
        cal = int(cal)
        if cal == 0:
            return spectrum
        if cal >= 0:
            spectrum = np.concatenate(([0]*cal,  spectrum[:-cal]))
        else:
            spectrum = np.concatenate((spectrum[-cal:], [0]*-cal))
        
        return spectrum
    
    
    def apply_origin_wavelength_calibration(self, spectrum, poly, origin):
        """applies a calibration to a spectrum around a specific wavelength
        params:
            spectrum: the spectrum to be calibrated
            poly: the polynomial being applied
            origin: the point around which the poynomial is being applied
        returns:
            calibrated: calibrated spectrum
        """
        left=spectrum[:origin]
        origin_value = np.array([spectrum[origin]])
        right = spectrum[origin+1:]
        
        ##### calibrates the spectrum to the left and right of the origin and adds the 3 together
        left_pos = int(SpectrumUtils.apply_polynomial(len(left), poly[0], poly[1], poly[2], poly[3]))
        cal_left = []
        if len(left) > 0:
            cal_left = self.apply_wavelength_calibration(left, poly, np.arange(len(left)), output_shape=[left_pos-len(left), left_pos], isdf=False)
        
        cal_right = []
        if len(right) > 0:
            cal_right = self.apply_wavelength_calibration(right, poly, np.arange(len(right)), output_shape=[0, len(right)], isdf=False)
        
        calibrated = np.concatenate((cal_left, origin_value, cal_right))
        return calibrated
    

    def apply_amplitude_calibration(self, spectrum, calibration):
        return spectrum*calibration


 # -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""

import os
import xml.etree.ElementTree as ET

import pandas as pd
import warnings

##### supressing pandas performance warnings
warnings.simplefilter(action='ignore', category=pd.errors.PerformanceWarning)


from hsr1.db import (
    SqliteDBLoad,
    SqliteDBStore,
    PreCalculations
)

from hsr1.utils.reformatData import ReformatData as reformat


class DBDriver:
    """Generic class for database operations.
    
    contains a load and a store object,
    also exposes common methods in those objects to the user for easy access
        e.g: driver.load rather than driver.db_load.load
    """
    def __init__(self, db_name:str, db_type:str="sqlite"):
        """
        params:
            db_name: name of the db. if the db dosent exist, it will be created
                when data is stored to it
        """
        self.db_name = db_name
        if self.db_name[-3:] != ".db":
            self.db_name += ".db"
        
        if db_type == "sqlite":
            self.db_store = SqliteDBStore(self)
            self.db_load = SqliteDBLoad(self.db_name)
        else:
            raise NotImplementedError("Only sqlite databases are currently supported")
    
    
    @property
    def last_timestamp(self) -> str:
        """returns a string of the last timestamp in the database, None if empty"""
        df = self.db_load.load_sql("SELECT pc_time_end_measurement FROM spectral_data ORDER BY pc_time_end_measurement DESC LIMIT 1")
        if df is None:
            return None
        return df.loc[0, 0] 
    
    @property
    def first_timestamp(self) -> str:
        """returns a string of the first timestamp in the database, None if empty"""
        df = self.db_load.load_sql("SELECT pc_time_end_measurement FROM spectral_data ORDER BY pc_time_end_measurement ASC LIMIT 1")
        if df is None:
            return None
        return df.loc[0, 0]
    
    ##### exposing some methods from store and load to make them easier to access
    ##### e.g: driver.load() rather than driver.load.load()
    def store(self, dfs:tuple, precalculate:bool=True):
        self.db_store.store(dfs, precalculate)
    
    def combine_database(self, new_db_name:str):
        self.db_store.combine_database(new_db_name)
    
    def combine_database_folder(self, folder, delete=False):
        self.db_store.combine_database_folder(folder, delete)
    
    def store_raw(self, dfs, deployment_metadata):
        self.db_store.store_raw(dfs, deployment_metadata)
    
    def load_table_names(self) -> pd.DataFrame:
        return self.db_load.load_table_names()
    
    def load_sql(self, sql:str) -> pd.DataFrame:
        return self.db_load.load_sql(sql)
    
    def load(self, columns:[str]=[], 
             table:str=None,
             start_time:str=None, 
             end_time:str=None,
             condition:str="",
             raise_on_missing:bool=True,
             sort:bool=True,
             deserialise:bool=True,
             timezone:str="+00:00"
             ) -> pd.DataFrame:
        return self.db_load.load(columns,
                                 table,
                                 start_time,
                                 end_time,
                                 condition,
                                 raise_on_missing,
                                 sort,
                                 deserialise,
                                 timezone)
    
    def load_accessory(self, columns:[str]=[], 
                       start_time:str=None, 
                       end_time:str=None,
                       condition:str="",
                       raise_on_missing:bool=True,
                       sort:bool=True,
                       timezone:str="+00:00"
                       ) -> pd.DataFrame:
        return self.db_load.load_accessory(columns,
                                           start_time,
                                           end_time,
                                           condition,
                                           raise_on_missing,
                                           sort,
                                           timezone)
    
    def load_metadata(self, columns:[str]=[], condition:str="",
                      raise_on_missing=True) -> pd.DataFrame:
        return self.db_load.load_metadata(columns, condition, raise_on_missing)
    
    def load_spectrum(self, column:str, 
             table:str=None,
             start_time:str=None, 
             end_time:str=None,
             condition:str="",
             raise_on_missing:bool=True,
             sort:bool=True,
             deserialise:bool=True,
             timezone:str="+00:00"
             ) -> pd.DataFrame:
        return self.db_load.load_spectrum(column,
                                          table,
                                          start_time,
                                          end_time,
                                          condition,
                                          raise_on_missing,
                                          sort,
                                          timezone)
    
    def load_raw(self, columns=[], start_time=None, end_time=None):
        return self.db_load.load_raw(columns, start_time, end_time)


    def exists(self) -> bool:
        ##### TODO: move somewhere, this is sqlite specific
        """checks if a database exists and has all columns, returns bool"""
        if not os.path.isfile(self.db_name):
            return False
        
        column_headings = self.db_load.load_table_names()
        if len(column_headings) > 0:
            return True
        else:
            return False
    
    
    def add_precalculated_values(self, method:str=None, mobile=True, sample_ids_to_add=None, drop_existing=True):
        """calculates some commonly used values and stores them to the database
        
        params:
            method: which library to use
                "ephem": basic calculation
                "sg2_static": really fast for a static instrument
                "sg2_mobile": slow, not reccomended, but works for a mobile dataset
                "sg2": detects wether the dataset is mobile and uses the appropriate method
                None(default): uses "sg2_static" if static, and ephem if mobile
            mobile: wether or not the dataset's location will change. if you know 
                it won't, you can pass mobile=False and calculations will be sped up
            sample_ids_to_add(np.array): array of sample ids that correspond to 
                the readings that precaclulated_values will be calculated for
            drop_existing: deletes the existing data in the database. default true
                as this function is usually used to recalculate a whole dataset.
                To fill in a gap or recalculate a small section and keep the rest,
                set this to False and pass in the sample ids you want to add
                
        no returns, just updates the database with a new table
        """
        ref = reformat()
        
        deployment_metadata = self.load_metadata()
        
        if drop_existing:
            self.db_store.drop_table("precalculated_values")
        
        p_calcs = PreCalculations(deployment_metadata=deployment_metadata)
        spectral_data = self.db_load.load(["pc_time_end_measurement", "sample_id", "dataseries_id"])
        

        data = None
        try:
            data = self.db_load.load(p_calcs.requirements+["sample_id"])
            print("loaded")
        except KeyError:
            data = pd.DataFrame()
            data["pc_time_end_measurement"] = spectral_data["pc_time_end_measurement"]
        
        if "gps_longitude" not in data.columns:
            data["gps_longitude"] = deployment_metadata["default_longitude"].iloc[0]
            data["gps_latitude"] = deployment_metadata["default_latitude"].iloc[0]
            data["gps_altitude"] = deployment_metadata["default_elevation"].iloc[0]
        
        
        if sample_ids_to_add is not None:
            select_data = data.set_index("sample_id")
            data_to_calculate = select_data.loc[sample_ids_to_add, :]
            data_to_calculate["sample_id"] = data_to_calculate.index
            data = data_to_calculate.reset_index(drop=True)
        
        precalculated_values = p_calcs.calculate_all(data, method)
        
        precalculated_values = reformat().reset_index(precalculated_values, "pc_time_end_measurement")
        precalculated_values["dataseries_id"] = spectral_data["dataseries_id"]
        precalculated_values["sample_id"]  = spectral_data["sample_id"]
        
        deployment_ids = {}
        timezones = {}
        timedeltas_series = precalculated_values["dataseries_id"].copy()
        for i in range(len(deployment_metadata)):
            deployment_id = deployment_metadata.loc[i, "deployment_id"]
            dataseries_id = deployment_metadata.loc[i, "dataseries_id"]
            deployment_ids[dataseries_id] = deployment_id
            
            timezone, timedelta = ref.calculate_timezone(deployment_metadata.loc[i, "timezone"])
            timezones[deployment_metadata.loc[i, "deployment_id"]] = timezone
            
            timedeltas_series[timedeltas_series == dataseries_id] = timedelta
        
        
        timezones_series = precalculated_values["dataseries_id"].replace(deployment_ids).replace(timezones)
        
        
        precalculated_values["pc_time_end_measurement"] = pd.to_datetime(precalculated_values["pc_time_end_measurement"], utc=True)
        precalculated_values["pc_time_end_measurement"] = (precalculated_values["pc_time_end_measurement"]+timedeltas_series).astype(str)
        precalculated_values["pc_time_end_measurement"] = precalculated_values["pc_time_end_measurement"].str.slice(stop=19)
        precalculated_values["pc_time_end_measurement"] = precalculated_values["pc_time_end_measurement"]+timezones_series
        
        ##### remove rows where data couldnt be calculated
        precalculated_values = precalculated_values.loc[~precalculated_values["sza"].isnull(), :]
        
        print("storing precalculated values")
        self.db_store.store_dataframe(precalculated_values, "precalculated_values")
    
    
    def make_gpx(self, filename:str):
        """make and save an .xml file
        
        can be used on google my maps or another gpx viewer
        
        params:
            filename: the file location that the .gpx file will be saved to
        
        """
        
        data = self.db_load.load(["pc_time_end_measurement", "gps_latitude", "gps_longitude", "gps_altitude"])
        
        ##### select the first reading from each hour
        data["pc_time_end_measurement"] = data["pc_time_end_measurement"].astype(str).str.slice(stop=13)
        data = data.ffill().bfill()
        data = data.groupby("pc_time_end_measurement").first()
        data = data.reset_index()
        data["pc_time_end_measurement"] = data["pc_time_end_measurement"].str.slice(stop=10) + "T" + data["pc_time_end_measurement"].str.slice(11) + ":00:00"
        
          
        ##### settings for the gpx(root) tag
        gpx = ET.Element("gpx")
        gpx.set("version", "1.1")
        gpx.set("creator", "Peak Design")
        
        ##### setting metadata
        metadata = ET.SubElement(gpx, "metadata")
        metadata.set("name", "hsr-1 location data")
        author = ET.SubElement(metadata, "author")
        author_name = ET.SubElement(author, "name")
        author_name.text = "Peak Design"
        
        ##### makes a track element that will contain all the track data
        trk = ET.SubElement(gpx, "trk")
        name = ET.SubElement(trk, "name")
        name.text = "HSR-1 gps data"
        trkseg = ET.SubElement(trk, "trkseg")
        
        for i in range(len(data)):
            ##### makes a new track point for each 
            trkpt = ET.SubElement(trkseg, "trkpt")
            trkpt.set("lat", str(data.loc[i, "gps_latitude"]))
            trkpt.set("lon", str(data.loc[i, "gps_longitude"]))
            trkpt.set("ele", str(data.loc[i, "gps_altitude"]))
            
            time = ET.SubElement(trkpt, "time")
            time.text = data.loc[i, "pc_time_end_measurement"]
        
        #####
        xml = ET.tostring(gpx)
        
        if filename[-4:] != ".gpx":
            filename += ".gpx"
        
        with open(filename, "wb") as f:
            f.write(xml)

    
            
            



# -*- coding: utf-8 -*-
"""
Created on Wed Feb 28 17:14:59 2024

@author: albie
"""

from hsr1.utils.spectrum.pixelSpectrum import PixelSpectrum
from hsr1.utils.spectrum.rawDataset import RawDataset
from hsr1.utils.spectrum.weightedSpectrum import WeightedSpectrum
from hsr1.utils.spectrum.weightedSpectrumSegment import WeightedSpectrumSegment
import hsr1.utils.spectrum.spectrumUtils as spectrumUtils


# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""
from datetime import datetime

import numpy as np


class LatLonGraph:
    def __init__(self, data):
        self.data = data
    
    
    def bin_data(self, data, bins=None):
        """bins a series of data
        generates a list of values across the dataset, and rounds each value to 
        the closest bin
        
        params:
            data: the dataset to be binned
            bins: the number of bins to be created
        returns:
            result: the original dataset, rounded into the new bins
            bin_centres: the locations of the bins
        """
        
        ##### generates the centre of the bins
        bin_centres = np.linspace(min(data), max(data), num=bins)
        
        ##### make a 2d array of the distance between every measured point and the centre of each bin
        distances = np.abs(np.subtract.outer(bin_centres, data.to_numpy()))
        
        ##### find the index of the closest bin
        indecies = np.argmin(distances, axis=0)
        
        ##### apply the index to match the bin to where the actual value was
        result = bin_centres[indecies]
        return result, bin_centres

    
    def plot_lat_lon(self, axes, column=None, stack_resolution="max", lat_lon_name=["Latitude", "Longitude"], bins=100, title=None):
        """plots a track of the location of the dataset, with another column as the colour
        params:
            axes: the axes the graph will be plotted on
            column: the name of the column that will be plotted as colour, None=density
            stack_resolution: what to do when multiple datapoints are in the same bin
                            options: "max", "min", "mean", "mode", "first", "count"
            lat_lon_name: what the latitude and longitude are called in the dataframe
            bins: how many bins to put the data into
        """
        if title is None:
            if column is None:
                title = "Location Density"
            else:
                title = column
        
        lat_name = lat_lon_name[0]
        lon_name = lat_lon_name[1]
        
        ##### generate the bins, if there are fewer unique datapoints than bins, use number of datapoints or spread with 0.1 increments, whichever is smaller
        #####   max is 100
        new_bins = [bins, bins]
        
        lat_unique = len(self.data[lat_name].unique())
        lon_unique = len(self.data[lon_name].unique())
        
        if lon_unique < bins:
            evenly_spaced_lon = int(max(((self.data[lon_name])-min(self.data[lon_name]))/0.01)+1)
            new_bins[0] = min(100, max(lon_unique, evenly_spaced_lon))
        if lat_unique < bins:
            evenly_spaced_lat = int(max(((self.data[lat_name])-min(self.data[lat_name]))/0.01)+1)
            new_bins[1] = min(100, max(lat_unique, evenly_spaced_lat))
        
        ##### handle None column name, column dosent matter, just stack_resolution=count
        if column is None:
            column="pc_time_end_measurement"
            stack_resolution = "count"
        
        ##### load correct data from the dataframe and process if needed
        if column == "pc_time_end_measurement" or column == "density":
            self.data["data_col"] = self.data["pc_time_end_measurement"].dt.tz_localize(None).astype("datetime64[ns]").astype("int64")
        else:
            self.data["data_col"] = self.data[column]
        
        
        longitude, lon_bins = self.bin_data(self.data[lon_name], bins=new_bins[0])
        latitude, lat_bins = self.bin_data(self.data[lat_name], bins=new_bins[1])
        
        ##### group the data according to the bins, aggregate according to stack_resolution
        if stack_resolution == "max":
            new_df = self.data["data_col"].groupby([longitude, latitude]).max().unstack(level=0).sort_index(ascending=False)
        elif stack_resolution == "min":
            new_df = self.data["data_col"].groupby([longitude, latitude]).min().unstack(level=0).sort_index(ascending=False)
        elif stack_resolution == "mean":
            new_df = self.data["data_col"].groupby([longitude, latitude]).mean().unstack(level=0).sort_index(ascending=False)
        elif stack_resolution == "mode":
            new_df = self.data["data_col"].groupby([longitude, latitude]).agg(lambda x: x.value_counts().index[0]).unstack(level=0).sort_index(ascending=False)
        elif stack_resolution == "first":
            new_df = self.data["data_col"].groupby([longitude, latitude]).first().unstack(level=0).sort_index(ascending=False)
        elif stack_resolution == "count":
            new_df = self.data["data_col"].groupby([longitude, latitude]).count().unstack(level=0).sort_index(ascending=False)
        
        
        new_df = new_df.reindex(index=np.flip(lat_bins), columns=lon_bins)
        
        ##### helper function to convert string timstamps to seconds since epoch
        def str_to_epoc_s(string):
            return datetime.strptime(string, "%Y-%m-%d").timestamp()
        
        
        ##### define im here for scope
        im = None
        ##### when data is stacked, set the colourmap to range over the actual data rather than the data in the image
        if stack_resolution in ["max", "min", "mean", "first"]:
            im = axes.imshow(new_df, cmap="jet", aspect="auto", 
                             # vmin=str_to_epoc_s(np.min(self.data["pc_time_end_measurement"]).date())*10**9,
                             # vmax=str_to_epoc_s(np.max(self.data["pc_time_end_measurement"]).date())*10**9)
                             vmin=np.min(self.data["pc_time_end_measurement"]).timestamp()*10**9,
                             vmax=np.max(self.data["pc_time_end_measurement"]).timestamp()*10**9)
        else:
            im = axes.imshow(new_df, cmap="jet", aspect="auto")
        axes.set_title(title)
        
        
        self.ticks(new_bins, axes, lon_name, lat_name)
        
        return im
        
    def ticks(self, bins, axes, lon_name, lat_name):
        """adds accurate lat and lon ticks to the axes"""
        
        ##### handle when there are fewer bins than ticks,
        ##### no point having more detail on the axes than on the graph
        num_ticks = [5, 5]
        if bins[0] < num_ticks[0]:
            num_ticks[0] = bins[0]
        if bins[1] < num_ticks[1]:
            num_ticks[1] = bins[1]
        
        lon_max = np.max(self.data[lon_name])
        lon_min = np.min(self.data[lon_name])
        lonlim = axes.get_xlim()
        ##### ticks are in the middle of the bins, not the edges
        lonlim = lonlim[0]+0.5, lonlim[1]-0.5
        
        xticks = np.linspace(lonlim[0], lonlim[1], num_ticks[0])
        xlabels = np.linspace(lon_min, lon_max, num_ticks[0]).round(4)
        axes.set_xticks(xticks, xlabels)
        
        lat_max = np.max(self.data[lat_name])
        lat_min = np.min(self.data[lat_name])
        latlim = axes.get_ylim()
        ##### ticks are in the middle of the bins, not the edges
        latlim = latlim[0]-0.5, latlim[1]+0.5
        
        yticks = np.linspace(latlim[0], latlim[1], num_ticks[1])
        ylabels = np.linspace(lat_min, lat_max, num_ticks[1]).round(4)
        axes.set_yticks(yticks, ylabels)


# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""

import matplotlib
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
import numpy as np
import pandas as pd

import time

import hsr1.plots.graphUtils as graphUtils

class ElvAziGraph:
    def __init__(self, cmap=plt.get_cmap("jet"), diffuse_name="dif"):
        self.diffuse_name = diffuse_name
        self.cmap = cmap
        ##### according to brsn limits, highest possible global value is 1407
        self.max_integral = None
        self.requirements = {"ghi":["pc_time_end_measurement", "sza", "azimuth", "global_integral"],
                             "diff":["pc_time_end_measurement", "sza", "azimuth", "diffuse_integral"],
                             "dni":["pc_time_end_measurement", "sza", "azimuth", "direct_normal_integral"],
                             "horizon_line":["gps_latitude", "gps_longitude", "gps_altitude"],
                             "all":["pc_time_end_measurement", "sza", "azimuth", "diffuse_integral", "global_integral", "direct_normal_integral", "gps_latitude", "gps_longitude", "gps_altitude"]}
        self.flags = None
    
    
    def __calc_horizon_line(self, location):
        """reads the horizon line data
        from: https://github.com/AssessingSolar
        params:
            location: [lat, lon, elv]
        """
        import uuid
        import os
        import pandas
        from urllib.request import urlopen

        if (len(location) == 2):
            location[3] = -999;

        #unique id
        uid = uuid.uuid4().hex

        fic_output_csv = 'horizon_srtm_output_{}.csv'.format(uid)
        str_wps = 'http://toolbox.webservice-energy.org/service/wps?service=WPS&request=Execute&identifier=compute_horizon_srtm&version=1.0.0&DataInputs=';
        datainputs_wps = 'latitude={:.6f};longitude={:.6f};altitude={:.1f}'\
    	.format(location.iloc[0], location.iloc[1], location.iloc[2]);
        
        response = urlopen('{}{}'.format(str_wps,datainputs_wps))
        HZ = pandas.read_csv(response,delimiter=';',comment='#',header=None,skiprows=16,nrows=360,names=['AZIMUTH', 'ELEVATION'])
        
        if HZ["ELEVATION"].iloc[0] == np.nan:
            raise Exception("could not calcuate horizon line")
        
        return HZ
        
    
    def elv_azi_graph(self, 
                      axes,
                      df:pd.DataFrame, 
                      column_name:str, 
                      show_xlabels:bool=True, 
                      show_ylabels:bool=True,
                      show_cbar:bool=True, 
                      show_horizon=False):
        """plots a graph of elevation against azimuth with the intensity shown by colour
        params:
            axes: the axes to be plotted onto
            df: the data to be plotted
            column_name: the name of the column containing the data to be plotted
            show_xlabels: whether or not to show the x axes labels. useful when plotting multiple of these graphs with the same labels
            show_cbar: whether or not to show the colorbar. useful when plotting multiple of these graphs with the same colormap
        """
        
        if self.max_integral is None:
            self.max_integral = np.max(df[column_name])
        
        
        bbox = axes.get_window_extent()
        vert_pixels = int(bbox.height)
        horiz_pixels = int(bbox.width)
        
        ##### copy to have local scope
        df = df.copy()
        df["elv"] = 90-np.degrees(df["sza"])
        df["azimuth"] = np.degrees(df["azimuth"])
        
        daytime = df["elv"] >= 0
        elv = np.round(df["elv"][daytime]).astype(int)
        azi = np.round(df["azimuth"][daytime]).astype(int)
        
        if len(elv) == 0:
            return None
        
        ##### filter out nighttime
        df = df.loc[df["elv"] >= 0]
        
        ##### rounds the data to the nearest int, selects the max irradiance value, and stores them as a 2D array
        new_df = df[column_name].groupby([df["azimuth"].round(0), df["elv"].round(0)]).max().unstack(level=0).sort_index(ascending=False)
        
        new_index = np.arange(0, max(elv)+1)
        new_cols = np.arange(min(azi), max(azi)+1)
        new_df = new_df.reindex(index=np.flip(new_index), columns=new_cols)
        
        
        extent = (min(azi), max(azi), 0, max(elv))
        im = axes.imshow(new_df, aspect="auto", interpolation="none", cmap="jet", extent=extent, vmin=0, vmax=self.max_integral)
        
        
        if self.flags is not None:
            colordict = {
                "red":((0,1,1), (1,1,1)),
                "green":((0,1,1), (1,0,0)),
                "blue":((0,1,1), (1,1,1))}
            flag_cmap = LinearSegmentedColormap("LinearCmap", colordict)
            
            any_flags = pd.DataFrame()
            any_flags["flags"] = self.flags.any(axis=1)
            
            new_df = any_flags["flags"].groupby([df["azimuth"].round(0), df["elv"].round(0)]).any().unstack(level=0).sort_index(ascending=False)
            
            new_index = np.arange(0, max(elv)+1)
            new_cols = np.arange(min(azi), max(azi)+1)
            new_df = new_df.reindex(index=np.flip(new_index), columns=new_cols)
            
            new_df = new_df.astype(float).fillna(0)
            
            ##### alpha controls the transparancy of the image, this sets alpha to be opaque when flagged, and transparent when not
            ##### removes noise on the original graph
            alpha = new_df.values.astype(float)
            im = axes.imshow(new_df, aspect="auto", norm=matplotlib.colors.LogNorm(vmin=0, vmax=1), alpha=alpha, interpolation="none", cmap=flag_cmap, extent=extent)
        
        
        ylim = axes.get_ylim()
        axes.set_ylim(ylim[0], ylim[1]*1.05)
        
        if show_ylabels:
            axes.set_ylabel("solar elevation(°)")
        
        xticks = np.linspace(extent[0], extent[1], 5)
        if show_xlabels:
            axes.set_xlabel("solar azimuth(°)")
            axes.set_xticks(xticks)
        else:
            axes.set_xticks(xticks, labels=[])
        
        if show_cbar:
            self.fig.colorbar(im, ax=axes, pad=0)
    
        if show_horizon:
            ##### plots the horizon line onto the graph
            try:
                horizon_line = self.__calc_horizon_line(df.loc[0, ["gps_latitude", "gps_longitude", "gps_altitude"]])
                if type(horizon_line["AZIMUTH"].iloc[0]) != type(""):
                    axes.plot(horizon_line["AZIMUTH"], horizon_line["ELEVATION"])
                else:
                    print("invalid horizon")
            except:
                print("horizon could not be calculated")
        
        
        return im
    
    
    def ghi(self, axes, df:pd.DataFrame, show_xlabels:bool=True, show_ylabels:bool=True, show_cbar=True, show_horizon=True):
        """graphs an elevation/azimuth plot for ghi
        params:
            axes: the axes to plot on
            df: the data to be plotted
            show_xlabels: wether or not to show labels on the x axis, useful when plotting multiple at once
            show_cbar: wether or not to show a colorbar, useful when plotting multiple at once
        """
        self.elv_azi_graph(axes, df, "global_integral", show_xlabels=show_xlabels, show_ylabels=show_ylabels, show_cbar=show_cbar, show_horizon=show_horizon)
        mirror_axes = axes.twinx()
        mirror_axes.set_yticks([])
        mirror_axes.set_ylabel("GHI(W/m²)")
        
        
    
    def diff(self, axes, df:pd.DataFrame, show_xlabels:bool=True, show_ylabels:bool=True, show_cbar=True, show_horizon=True):
        """graphs an elevation/azimuth plot for diffuse
        params:
            axes: the axes to plot on
            df: the data to be plotted
            show_xlabels: wether or not to show labels on the x axis, useful when plotting multiple at once
            show_cbar: wether or not to show a colorbar, useful when plotting multiple at once
        """
        self.elv_azi_graph(axes, df, "diffuse_integral", show_xlabels=show_xlabels, show_ylabels=show_ylabels, show_cbar=show_cbar, show_horizon=show_horizon)
        mirror_axes = axes.twinx()
        mirror_axes.set_yticks([])
        mirror_axes.set_ylabel(self.diffuse_name+"(W/m²)")
        
    
    def dni(self, axes, df:pd.DataFrame, show_xlabels:bool=True, show_ylabels:bool=True, show_cbar=True, show_horizon=True): 
        """graphs an elevation/azimuth plot for dni
        params:
            axes: the axes to plot on
            df: the data to be plotted
            show_xlabels: wether or not to show labels on the x axis, useful when plotting multiple at once
            show_cbar: wether or not to show a colorbar, useful when plotting multiple at once
        """
        self.elv_azi_graph(axes, df, "direct_normal_integral", show_xlabels=show_xlabels, show_ylabels=show_ylabels, show_cbar=show_cbar, show_horizon=show_horizon)
        mirror_axes = axes.twinx()
        mirror_axes.set_yticks([])
        mirror_axes.set_ylabel("DNI(W/m²)")
        
    
    def graph_all(self, axes, df:pd.DataFrame, fig, flags=None, show_horizon=True):
        """plots all the possible graphs of this type
        params:
            axes: list of axes that the graphs will be plotted on
            df: the data to be plotted
            fig: the figure all the graphs are plotted on, for plotting a shared colorbar
        """
        self.fig = fig
        self.flags = flags
        hz_requirements = self.requirements["horizon_line"]
        self.ghi(axes[0], df[self.requirements["ghi"]+hz_requirements], show_xlabels=False, show_ylabels=False, show_cbar=False, show_horizon=show_horizon)
        self.diff(axes[1], df[self.requirements["diff"]+hz_requirements], show_xlabels=False, show_ylabels=True, show_cbar=False, show_horizon=show_horizon)
        self.dni(axes[2], df[self.requirements["dni"]+hz_requirements], show_xlabels=True, show_ylabels=False, show_cbar=False, show_horizon=show_horizon)
        
        graphUtils.cbar(axes, fig, self.max_integral, 100)
        


# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""
import datetime

import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
import pandas as pd
import numpy as np

from hsr1.plots import graphUtils

import time

class LinearTimeGraph:
    def __init__(self, timezone=None, diffuse_name="dif"):
        self.diffuse_name = diffuse_name
        self.timezone = timezone
        ##### according to brsn limits, highest possible global value is 1407
        self.max_integral = 1407
        self.requirements = {"all":["pc_time_end_measurement", "global_integral", "diffuse_integral", "direct_normal_integral"],
                             "ghi":["pc_time_end_measurement", "global_integral"],
                             "diff":["pc_time_end_measurement", "diffuse_integral"],
                             "dni":["pc_time_end_measurement", "direct_normal_integral"]}
        
        self.flags = None
    
    
    def linear_integral_plot_all(self, df, axes, time_col, data_col, show_xlabels=True, date_gap=None, date_gap_scale=1000):
        """plots the intensity per minute (or per several minutes)
        params:
            df: the data to be plotted
            axes: the axes to be plotted on
            time_col: the name of the column containing time data
            data_col: the name of the column containing intensity data
            show_xlabels: whether or not to show the x axes labels. useful when plotting multiple of these graphs with the same labels
            date_gap: the proportion of readings to include, 1/min gets very slow
            date_gap_scale: when no date_gap is specified, this can generate a date gap value based on the size of the dataset
        """
        df = df.copy()
        
        df[time_col] = pd.DatetimeIndex(df[time_col]).tz_localize(None)
        
        if min(df[data_col]) == max(df[data_col]):
            print("min value is the same as max value")
            return
        
        max_height = round(max(df[data_col]))
        max_height = round(self.max_integral)
        
        bbox = axes.get_window_extent()
        horizontal_pixels = int(bbox.width)
        
        
        colordict = {
            "red":((0,1,1), (1,0,0)),
            "green":((0,1,1), (1,0,0)),
            "blue":((0,1,1), (1,1,1))}
        cmap = LinearSegmentedColormap("LinearCmap", colordict)
        
        if date_gap == None:
            date_gap = len(df)//date_gap_scale
            date_gap = max(date_gap, 1)
        
        
        
        measured_timestamps = df["pc_time_end_measurement"]
        first_reading_int = np.datetime64(str(measured_timestamps.iloc[0].date())+" 00:00:00").astype("int64")*10**9
        last_reading_int = np.datetime64(str(measured_timestamps.iloc[len(df)-1].date())+" 23:59:59").astype("int64")*10**9
        measured_timestamps = measured_timestamps.values.astype("int64")
        
        diff = np.diff(measured_timestamps)
        min_diff = np.min(diff[diff>0])
        num_groups = round((last_reading_int-first_reading_int)/min_diff)
        pixels_to_use = min(num_groups, horizontal_pixels)
        
        pixel_timestamps = np.linspace(first_reading_int, last_reading_int, pixels_to_use)
        
        
        readings_df = pd.DataFrame()
        readings_df[time_col] = pd.DatetimeIndex(pixel_timestamps)
        if self.flags is not None:
            df["flags"] = self.flags.any(axis=1)
        
        readings_df = readings_df.merge(df, on=time_col, how="outer").replace(np.nan, 0)
        
        readings_df["int_time"] = readings_df[time_col].astype("int64")
        
        
        def closest_bin(i):
            timestamp = readings_df.loc[i, "int_time"]
            return pixel_timestamps[np.argmin(np.abs(pixel_timestamps-timestamp))]
        
        readings_df = readings_df.groupby(closest_bin).max()
        
        
        data = [np.concatenate((np.zeros(max_height-max(0, round(x))), np.ones(max(0,round(x))))) for x in readings_df[data_col].clip(upper=self.max_integral)]
        data = np.array(data).T
        
        extent = (min(df[time_col]).timestamp(), max(df[time_col]).timestamp(), 0, self.max_integral)
        
        
        axes.imshow(data, aspect="auto", interpolation="none", vmin=0, cmap=cmap, extent=extent)
        
        
        if self.flags is not None:
            colordict = {
                "red":((0,1,1), (1,1,1)),
                "green":((0,1,1), (1,0,0)),
                "blue":((0,1,1), (1,1,1))}
            flag_cmap = LinearSegmentedColormap("LinearCmap", colordict)
            
            any_flags = pd.DataFrame()
            any_flags["flags"] = readings_df["flags"]
            
            flagged = readings_df["flags"].values
            
            flagged_data = data*(flagged.T)
            
            axes.imshow(data, alpha=flagged_data.astype(float), aspect="auto", interpolation="none", vmin=0, cmap=flag_cmap, extent=extent)
        
        
        xticks, labels = graphUtils.calculate_date_labels(df[time_col], axes.get_xlim())
        if not show_xlabels:
            labels = []
        
        axes.set_xticks(xticks, labels)
    
    
    def ghi(self, axes, df, show_xlabels=True, timezone=None):
        """graphs a linear time plot for ghi
        params:
            axes: the axes to plot on
            df: the data to be plotted
            show_xlabels: wether or not to show labels on the x axis, useful when plotting multiple at once
        """
        self.linear_integral_plot_all(df, axes, "pc_time_end_measurement", "global_integral", show_xlabels)
        axes.set_ylabel("GHI(W/m²)")
    
    def diff(self, axes, df, show_xlabels=True, timezone=None):
        """graphs a linear time plot for ghi
        params:
            axes: the axes to plot on
            df: the data to be plotted
            show_xlabels: wether or not to show labels on the x axis, useful when plotting multiple at once
        """
        self.linear_integral_plot_all(df, axes, "pc_time_end_measurement", "diffuse_integral", show_xlabels)
        axes.set_ylabel(self.diffuse_name+"(W/m²)")
    
    def dni(self, axes, df, show_xlabels=True, timezone=None):
        """graphs a linear time plot for ghi
        params:
            axes: the axes to plot on
            df: the data to be plotted
            show_xlabels: wether or not to show labels on the x axis, useful when plotting multiple at once
        """
        dfa = df.copy()
        dfa["direct_normal_integral"] = dfa["direct_normal_integral"].fillna(0)
        self.linear_integral_plot_all(dfa, axes, "pc_time_end_measurement", "direct_normal_integral", show_xlabels)
        axes.set_ylabel("DNI(W/m²)")
        
    
    
    def graph_all(self, axes, df, flags=None):
        """plots all the possible graphs of this type
        params:
            axes: the axes to be plotted on
        """
        self.flags = flags
        
        self.ghi(axes[0], df[self.requirements["ghi"]], False)
        self.diff(axes[1], df[self.requirements["diff"]], False)
        self.dni(axes[2], df[self.requirements["dni"]], True)


# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""
import importlib_resources

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from hsr1.utils.spectrum import spectrumUtils as SpectrumUtils
from hsr1.utils.spectrum.pixelSpectrum import PixelSpectrum
from hsr1.utils.spectrum.weightedSpectrum import WeightedSpectrum


class RawDataset:
    def __init__(self, dataset:pd.DataFrame, 
                 calibrations:dict=None, calibrations_filepath=None, 
                 amplitude_calibration=None, amplitude_calibration_filepath=None, 
                 spectrum_type="wavelength"):
        """stores a raw dataset containing all the channels of raw data
        
        params:
            dataset: dataframe with columns "pc_time_end_measurement", "channel_0", "channel_1" etc.
            calibrations: dict containing polynomials with keys "channel_0", "channel_1" etc.
            calibrations_filepath: filepath to the current wavelength calibration file
            amplitude_cal: an amplitude calibration to apply to found spectrum, only used for graphs
            amplitude_calibration_filepath
        """
        self.dataset = dataset
        self.calibrations = calibrations
        self.amplitude_calibration = amplitude_calibration
        self.spectrum_type = spectrum_type
        
        if spectrum_type != "wavelength" and spectrum_type != "pixel":
            raise ValueError("spectrum_type must be either \"wavelength\" or \"pixel\"")
        
        if calibrations is None and spectrum_type == "wavelength":
            if calibrations_filepath is None:
                raise ValueError("""original calibrations must be passed, either as a dict or as the location of the calibration file. If this is uncalibrated data, pass spectrum_type=\"pixel\"""")
            self.calibrations = SpectrumUtils.read_wavelength_polys(calibrations_filepath)
        
        if spectrum_type == "pixel":
            if calibrations is not None or calibrations_filepath is not None:
                print("there is no point passing existing calibrations to a pixel dataset, pixel readings are assumed to be uncalibrated. \
                      \n if this is a calibration that is used on this spectrometer, you can change the default_calibration parameter in find_all_calibrations.")
        
        if amplitude_calibration is None and amplitude_calibration_filepath is not None:
            self.amplitude_calibration = SpectrumUtils.read_amplitude_file(amplitude_calibration_filepath)
        
        if amplitude_calibration is None and amplitude_calibration_filepath is None:
            self.amplitude_calibration = np.ones(801)
        
    
    
    def plot_calibration(self, polys, names=None):
        """plots a line graph of pixel against wavelength for each channel
        params:
            polys: list of lists or dict containing the polynommials for each channel's new calibration
            names: list of names to display on the legend
        """
        ps = PixelSpectrum()
        
        
        if isinstance(polys, dict):
            poly_names = []
            poly_list = []
            for poly in polys:
                poly_list.append(polys[poly])
                poly_names.append(poly)
            polys = poly_list 
            names = poly_names
        
        if names is None:
            names = [""]*len(polys)
        
        plt.figure(figsize=(20,10))
        for poly, name in zip(polys, names):
            plt.plot(ps.apply_wavelength_calibration(np.arange(0, 1200), poly, inverse=True), np.arange(300, 1101), linewidth=0.5, label=name + ": "+SpectrumUtils.poly_to_string(poly))
        plt.xlim(0, 1200)
        plt.suptitle("individual channel calibrations")
        plt.xlabel("pixels")
        plt.ylabel("wavelength")
        plt.legend()
        plt.show()
    
    def plot_calibration_wide(self, polys, names=None, title=""):
        """plots a line graph of pixel against wavelength for each channel
        params:
            polys: list of lists containing the polynommials for each channel's new calibration
            names: list of names to display on the legend
        """
        if names is None:
            names = [""]*len(polys)
        
        plt.figure(figsize=(20,10))
        for poly, name in zip(polys, names):
            x = np.linspace(-10000, 10000, 1000)
            y = [SpectrumUtils.apply_polynomial(i, poly[0], poly[1], poly[2], poly[3]) for i in x]
            plt.plot(x, y, label=name + ": " + SpectrumUtils.poly_to_string(poly))

        plt.axvline(0)
        plt.axvline(1200)
        plt.suptitle(title)
        plt.legend()
        plt.show()
    
    def plot_all_errors(self, polys, dips, names):
        plt.figure(figsize=(20, 10))
        for i, poly, _dips, name in zip(list(range(len(polys))), polys, dips, names):
            measured = _dips[0]
            ref = _dips[1]
            error = measured-ref
            
            plt.plot(ref, error-(i/50), label=name)
        
        plt.legend()
        plt.show()
    
    
    def find_all_calibrations(self, file_output=None, 
                              reference_spectrum=None, reference_filepath=None, 
                              plot_total_output=True, plot_individual_results=False, plot_debug=False, 
                              **kwargs):
        """loops through each channel in the dataset and finds the correct calibration
        params:
            file_output: for outputting results, which text file to use. leave as None to not output
            reference_spectrum: a spectrum in wavelength space that the measured data will be matched to
            plot_total_output: plots the calibration for each channel on one graph
            plot_individual_results: plots the results from each channel, passes plot_results to each calibration
            plot_debug: plots the result from each segment of each channel, passed to each calibration
            kwargs: arguments to pass to each channel's find_calibration method, cleaner than having all arguments as parameters here
        """
        if reference_filepath is None:
            res = importlib_resources.files("hsr1.data").joinpath("smarts.txt")
            file = importlib_resources.as_file(res)
            with file as f:
                reference_filepath = f
            
        
        
        if reference_spectrum is None:
            try:
                reference_spectrum = SpectrumUtils.read_simple_file(reference_filepath)
            except ValueError:
                reference_spectrum = SpectrumUtils.read_reference_file(reference_filepath)
        
        
        all_dips = []
        polys = []
        names = []
        new_calibrations = {}
        for channel_name in self.dataset.columns[1:]:
            
            
            ##### extract the polynomial that has already been applied to this channel
            applied_poly = None
            if self.spectrum_type == "wavelength":
                if channel_name in self.calibrations:
                    applied_poly = self.calibrations[channel_name]
                else:
                    print("error while calibrating "+channel_name+": couldnt find an existing calibration for this channel")
                    continue
            
            ps = PixelSpectrum(cal_poly=applied_poly)
            
            channel = self.dataset[["pc_time_end_measurement", channel_name]].copy()
            channel = channel.rename(columns={channel_name: "global_spectrum"})
            
            
            
            ##### calculate one pixel spectrum, for graphing
            pixel_spectrum = None
            if self.spectrum_type == "wavelength":
                pixel_data = ps.calculate_from_wavelengths(channel, ["global_spectrum"])
                max_index = 0
                if len(channel) > 1:
                    integrals = pixel_data["global_spectrum"].apply(np.sum)
                    max_index = np.argmax(integrals)
                pixel_spectrum = pixel_data.loc[max_index, "global_spectrum"]
            elif self.spectrum_type == "pixel":
                max_index = 0
                if len(channel) > 1:
                    integrals = channel.apply(np.sum)
                    max_index = np.argmax(integrals)
                pixel_spectrum = channel.loc[max_index, "global_spectrum"]
            
            if max(pixel_spectrum) == 0:
                print("invalid spectrum selected for graphing")
            
            ##### calculate the weighted spectrum and calibrate
            ws = None
            if self.spectrum_type == "wavelength":
                ws = WeightedSpectrum(wavelength_spectra=channel, reference_spectrum=reference_spectrum, 
                                      applied_poly=applied_poly, 
                                      filter_low_values=False, 
                                      amplitude_calibration=self.amplitude_calibration, 
                                      pixel_spectrum=pixel_spectrum, 
                                      name=channel_name, **kwargs)
            elif self.spectrum_type == "pixel":
                ws = WeightedSpectrum(pixel_spectra=channel, reference_spectrum=reference_spectrum, 
                                      applied_poly=applied_poly, 
                                      filter_low_values=False, 
                                      amplitude_calibration=self.amplitude_calibration, 
                                      pixel_spectrum=pixel_spectrum, 
                                      name=channel_name, **kwargs)
                
            wavelength_cal, dips = (None, None)
            if plot_total_output:
                wavelength_cal, dips = ws.find_calibration(plot_debug=plot_debug, plot_results=plot_individual_results, 
                                                            return_dips=plot_total_output, 
                                                            **kwargs)
            else:
                wavelength_cal = ws.find_calibration(plot_debug=plot_debug, plot_results=plot_individual_results, 
                                                            return_dips=plot_total_output, 
                                                            **kwargs)
            if wavelength_cal is not None:
                all_dips.append(dips)
                new_calibrations[channel_name] = wavelength_cal
                polys.append(wavelength_cal)
                names.append(channel_name)
        
        if plot_total_output:
            self.plot_calibration(polys, names)
            self.plot_calibration_wide(polys, names)
            self.plot_all_errors(polys, all_dips, names)
        
        if file_output is not None:
            SpectrumUtils.format_output(new_calibrations, file_output)
        
        return new_calibrations


# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""
import time

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

import math

sg2_imported = False
try:
    import sg2
    sg2_imported = True
except Exception as e:
    print(e)

import ephem

import hsr1.utils.HSRFunc as HsrFunc


class PreCalculations:
    def __init__(self, deployment_metadata):
        self.deployment_metadata=deployment_metadata
        
        self.requirements = ["gps_longitude", "gps_latitude", "gps_altitude", "pressure", "baro_temp", "pc_time_end_measurement", "dataseries_id"]
        self.sun_data = None
        self.sg2_values = ["topoc.gamma_S0", "topoc.alpha_S", "topoc.toa_hi", "geoc.R"]
    
    
    
    
    def __calculate_raw_sg2_static(self, data):
        location = [data[col].mean() for col in ["gps_longitude", "gps_latitude", "gps_altitude"]]
        
        datetimes = data.index.tz_localize(None).astype("datetime64[ns]")
        
        sun_position = sg2.sun_position([location], datetimes, self.sg2_values)
        
        result = {}
        result["elevation"] = sun_position.topoc.gamma_S0[0]
        result["azimuth"] = sun_position.topoc.alpha_S[0]
        result["toa_hi"] = sun_position.topoc.toa_hi[0]
        result["sed"] = sun_position.geoc.R
        
        result["elevation"] = self.atmospheric_correction(result["elevation"])
        
        return result
        
    
    
    def __calculate_raw_sg2_mobile(self, data):        
        locations = data[["gps_longitude", "gps_latitude", "gps_altitude"]].values

        sun_position = [sg2.sun_position([locations[i]], [pd.to_datetime(data.index[i]).to_datetime64()], self.sg2_values) for i in range(len(data))]
        
        result = {}
        result["elevation"] = [sun.topoc.gamma_S0[0][0] for sun in sun_position]
        result["azimuth"] = [sun.topoc.alpha_S[0][0] for sun in sun_position]
        result["toa_hi"] = [sun.topoc.toa_hi[0][0] for sun in sun_position]
        result["sed"] = [sun.geoc.R[0] for sun in sun_position]
        
        result["elevation"] = self.atmospheric_correction(result["elevation"])
        
        return result
    
    def __calculate_raw_ephem(self, data):
        instrument = ephem.Observer()
        sun = ephem.Sun()
        static_sun = ephem.Sun()
        
        locations = data[["gps_longitude", "gps_latitude", "gps_altitude"]].values
        dates = data.index.tz_localize(None)
        
        result = {}
        result["elevation"] = []
        result["azimuth"] = []
        result["sed"] = []
        
        for i in range(len(data)):
            current_loc = locations[i]
            
            instrument.lon = str(current_loc[0])
            instrument.lat = str(current_loc[1])
            instrument.elevation = current_loc[2]
            instrument.date = str(dates[i])
            
            sun.compute(instrument)
            
            static_sun.compute(str(dates[i]))
            
            result["elevation"].append(sun.alt)
            result["azimuth"].append(sun.az)
            result["sed"].append(static_sun.earth_distance)
        
        result_df = pd.DataFrame(result)
        
        day_times = result_df["elevation"] > 0
        solar_constant = 1367
        toa_ni = solar_constant/(result_df["sed"]**2)
        result_df["toa_hi"] = 0.0
        result_df.loc[day_times, "toa_hi"] = toa_ni[day_times]*np.sin(result_df.loc[day_times, "elevation"])
        
        result["toa_hi"] = result_df["toa_hi"].to_numpy()
        
        return result
            

    
    def calculate_all(self, data, method=None):
        """calculates all the precalculated values

        returns a dataframe with them all in as columns
        """
        
        
        if method is None:
            method = "ephem"
        
        if method in ["sg2", "sg2_static"] and not sg2_imported:
            print("sg2 was not sucessfully imported, using ephem")
            method = "ephem"
        
        
        ##### use sg2 static unless sg2 mobile is specified
        if method == "sg2":
            method = "sg2_static"
        
        data = data.set_index("pc_time_end_measurement")
        data = data.replace({None:"nan"})
        data = data.astype(str)
        data = data.replace({"":"nan"})
        
        float_cols = [col for col in data.columns if col not in ["dataseries_id", "sample_id"]]
        data[float_cols] = data[float_cols].astype(float)
        
        gps_col_dict = {"default_latitude":"gps_latitude", 
                        "default_longitude":"gps_longitude",
                        "default_elevation":"gps_altitude"}
        
        for idx in self.deployment_metadata.index:
            deployment_row = self.deployment_metadata.loc[idx, :]
            this_dataseries = data["dataseries_id"]==deployment_row["dataseries_id"]
            for col in gps_col_dict:
                sys_col = gps_col_dict[col]
                if deployment_row[col] != "''":
                    nan_values = np.logical_and(this_dataseries, data[sys_col].isnull())
                    data.loc[nan_values, sys_col] = float(deployment_row[col])
                
            
        
        self.data = data
        
        if method == "sg2_mobile":
            self.sun_data = self.__calculate_raw_sg2_mobile(data)
        elif method == "ephem":
            self.sun_data = self.__calculate_raw_ephem(data)
        elif method == "sg2_static":
            self.sun_data = self.__calculate_raw_sg2_static(data)
        else:
            raise ValueError(method+"is an invalid library to use")
        
        
        
        
        
        self.precalculated_values = pd.DataFrame()
        self.precalculated_values.index = self.data.index
        
        self.precalculated_values["sza"] = self.sza(data)
        self.precalculated_values["azimuth"] = self.azimuth()
        self.precalculated_values["toa_hi"] = self.top_of_atmosphere_hi()
        self.precalculated_values["sed"] = self.sun_earth_distance()
        self.precalculated_values["airmass"] = self.airmass(data)
        
        
        
        return self.precalculated_values
    
    
    def atmospheric_correction(self, elv):
        """corrects a list of elevation values for atmospheric refration
        converts from absolute position to apparent position
        """
        
        if not sg2_imported:
            raise ImportError("SG2 was not successfully imported")
        
        pressure = 1013.25
        temperature = 25
        
        elv_corr = sg2.topocentric_correction_refraction_SAE(elv, pressure, temperature)
        
        return elv_corr
    
    
    def sza(self, df):
        """unpacks the elevation and corrects it for atmospheric refraction
        returns: nparray of solar zenith angles
        
        uses pressure and temperature values from internal sensors, forward fills if there are missing values
        """
        # elv = np.array(self.sun_data["elevation"])
        
        # ##### puts all the relevant data into one dataframe to match timestamps, and for simplicity
        # data = pd.DataFrame()
        # data.index = df.index
        # data["elv"] = elv
        
        # data["pressure"] = 1.01325
        # data["temperature"] = 21
        # try:
        #     data["temperature"] = df["baro_temp"]+5.5
        # except:
        #     pass
        # data = data.ffill(axis=1)
        
        # ##### correct for atmospheric refraction
        # # elv_corr = sg2.topocentric_correction_refraction_SAE(data["elv"], data["pressure"], data["temperature"])
        # elv_corr = data["elv"]
        
        sza = (math.pi/2) - np.array(self.sun_data["elevation"])
        return sza
    
    def azimuth(self):
        """unpacks the azimuth value from sg2"""
        azi = self.sun_data["azimuth"]
        return azi
    
    def top_of_atmosphere_hi(self):
        """unpacks the top of atmosphere value from sg2"""
        toa = np.array(self.sun_data["toa_hi"])
        ##### adjusted because hsr1 only measures part of the spectrum
        return toa*(844/1367)
    
    def sun_earth_distance(self):
        """unpacks the sun earth distance value from sg2"""
        sed = self.sun_data["sed"]
        return sed
    
    def atmospheric_distance(self):
        """calculates the height of atmosphere that sun will travel through
        unneccessary, as earth is so much bigger than the atmosphere, this approaches parralel lines,
        can just use 1/cos(sza)
        """
        ##### TODO: use hsr_func's airmass function
            
        ##### radius of earth
        Re = 6371000
        
        ##### radius of atmosphere
        ##### very variable depending on what you want, should it be just as a ratio to atmospheric height?
        ##### if so then just set Ra to 1
        Ra = 1
        
        # solar zenith angle
        sza = self.precalculated_values["sza"]
        
        const = math.pow(Re, 2) + 2*Re*Ra + math.pow(Ra, 2)
        
        atmospheric_distance = np.sqrt(const - np.power(Re*np.sin(sza), 2)) - Re*np.cos(sza)
        
        return atmospheric_distance
    
    def cos_atmospheric_distance(self):
        Ra = 1
        sza = self.precalculated_values["sza"]
        return Ra/np.cos(sza)
        
    def airmass(self, df):
        data = pd.DataFrame()
        data["pressure"] = 1.01325
        data["sza"] = self.precalculated_values["sza"]
        daytimes = data["sza"] < 90
        data["airmass"] = np.nan
        data.loc[daytimes, "airmass"] = [HsrFunc.calc_air_mass(row[1]["sza"]) for row in data[daytimes].iterrows()]
    
        return data["airmass"]



































# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""
import pandas as pd
import numpy as np

all_columns = ["global_integral", "diffuse_integral", "direct_normal_integral",
               "sza", "toa_hi"]
    
def flag(df:pd.DataFrame, ignore_nights:bool=True):
    """generates a dataframe of each reading and wether that reading is flagged in each specific check
    params:
        df: dataframe columns: ["global_integral", "diffuse_integral", "direct_normal_integral", "sza", "toa_hi"]
        ignore_nights: if true, dosen't flag night data(sza<90)
    """
    limits = pd.DataFrame(index=df.index)
    flags = pd.DataFrame(index=df.index)
    
    columns = df.columns
    if "sza" in columns:
        limits["mu0"] = np.cos(df["sza"])
        if "toa_hi" in columns:
            limits["toa_ni"] = df["toa_hi"]/limits["mu0"]
    
    if np.isin(["sza", "toa_hi", "global_integral"], columns).all():
        limits["ghi_possible_min"] = -4
        limits["ghi_possible_max"] = limits["toa_ni"] * 1.5 * np.power(limits["mu0"], 1.2) + 100
        limits["ghi_rare_min"] = -2
        limits["ghi_rare_max"] = limits["toa_ni"] * 1.2 * np.power(limits["mu0"], 1.2) + 50
        flags["ghi_possible_min"] = df["global_integral"] < limits["ghi_possible_min"]
        flags["ghi_possible_max"] = df["global_integral"] > limits["ghi_possible_max"]
        flags["ghi_rare_min"] = df["global_integral"] < limits["ghi_rare_min"]
        flags["ghi_rare_max"] = df["global_integral"] > limits["ghi_rare_max"]
    
    if np.isin(["sza", "toa_hi", "diffuse_integral"], columns).all():
        limits["dif_possible_min"] = -4
        limits["dif_possible_max"] = limits["toa_ni"] *0.95 * limits["mu0"]**1.2 + 50
        limits["dif_rare_min"] = -2
        limits["dif_rare_max"] = limits["toa_ni"] *0.75 * limits["mu0"]**1.2 + 30
        flags["dif_possible_min"] = df["diffuse_integral"] < limits["dif_possible_min"]
        flags["dif_possible_max"] = df["diffuse_integral"] > limits["dif_possible_max"]
        flags["dif_rare_min"] = df["diffuse_integral"] < limits["dif_rare_min"]
        flags["dif_rare_max"] = df["diffuse_integral"] > limits["dif_rare_max"]
    
    if np.isin(["sza", "toa_hi", "direct_normal_integral"], columns).all():
        limits["dni_possible_min"] = -4
        limits["dni_possible_max"] = limits["toa_ni"]
        limits["dni_rare_min"] = -2
        limits["dni_rare_max"] = limits["toa_ni"] * 0.95 * limits["mu0"]**0.2 + 10
        flags["dni_possible_min"] = df["direct_normal_integral"] < limits["dni_possible_min"]
        flags["dni_possible_max"] = df["direct_normal_integral"] > limits["dni_possible_max"]
        flags["dni_rare_min"] = df["direct_normal_integral"] < limits["dni_rare_min"]
        flags["dni_rare_max"] = df["direct_normal_integral"] > limits["dni_rare_max"]
    
    if np.isin(["diffuse_integral", "global_integral"], columns).all():
        limits["diffuse_ratio"] = df["diffuse_integral"]/df["global_integral"]
        flags["diffuse_ratio_sza_possible_max"] = limits["diffuse_ratio"] > 1
    
    if np.isin(["global_integral", "sza", "toa_hi"], columns).all():
        limits["ghi_toa"] = df["global_integral"]/df["toa_hi"]
    
    
    if np.isin(["direct_normal_integral", "global_integral", "sza", "toa_hi"], columns).all():
        limits["dni_toani"] = df["direct_normal_integral"]/limits["toa_ni"]
        flags["dni_toani_ghi_toa_possible_max"] = (limits["dni_toani"] > 0.95) | (limits["ghi_toa"] > 1.35) | ((limits["dni_toani"] < 0.95) & (limits["dni_toani"] > limits["ghi_toa"]))
    
    if np.isin(["diffuse_integral", "global_integral", "sza", "toa_hi"], columns).all():
        flags["diffuse_ratio_ghi_toa"] = (limits["diffuse_ratio"] > 1) | ((limits["ghi_toa"] > 0.6) & (limits["diffuse_ratio"] > 0.95))
    
    if ignore_nights and "sza" in columns:
        flags.loc[df["sza"] > np.radians(90)] = False
    
    return flags






def calculate_limits():
    """generates evenly spaced szas and calculates the bsrn limits, assuming 1 sed. used as a reference on graphs"""
    limits = pd.DataFrame()
    solar_cost = 844
    limits["SZA"] = np.radians(np.arange(0,100,1))
    limits["Sa"] = solar_cost#/np.power(df["sed"], 2)
    limits["toa_hi"] = limits["Sa"]*np.cos(limits["SZA"])
    
    limits["mu0"] = np.cos(limits["SZA"])
    
    limits["ghi_possible"] = limits["Sa"]*1.5*np.power(limits["mu0"], 1.2) + 100
    limits["ghi_rare"] = limits["Sa"]*1.2*np.power(limits["mu0"], 1.2) + 50
    
    limits["diff_possible"] = limits["Sa"]*0.95*np.power(limits["mu0"], 1.2) + 50
    limits["diff_rare"] = limits["Sa"]*0.75*np.power(limits["mu0"], 1.2) + 30
    
    limits["dni_possible"] = limits["Sa"]
    limits["dni_rare"] = limits["Sa"]*0.95*np.power(limits["mu0"], 0.2) + 10
    
    limits["diff/ghi_sza_comparison"] = 1
    
    
    limits["ghi_toa"] = (np.arange(0,100,1)/100)*1.5
    limits["ghi_toa"] = limits["ghi_toa"].clip(upper=1.35)
    limits["dni_toani"] = limits["ghi_toa"]
    limits["dni_toani"] = limits["dni_toani"].clip(upper=0.95)
    
    limits["ghi/toa"] = (np.arange(0,100,1)/100)*1.5
    limits["diff/ghi"] = 1.1
    limits.loc[limits["ghi/toa"] > 0.6, "diff/ghi"] = 0.95
    
    
    
    return limits


# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""
import pandas as pd
import os
import zipfile

import numpy as np

import hsr1.read_txt.ImportHSRFiles as ImportHSRFiles
from hsr1.utils import (HSRFunc as hsr_func,
                       reformatData)

from hsr1 import DBDriver




def read(hsr_path="", start_date="2000-01-01", end_date="2100-01-01", deployment_metadata_filepath=None):
    """reads all the commonly used data from the .txt files. only loads the global and diffuse spectrum
    params:
        hsr_path: the filepath to the data that will be read. 
            this folder should contain one .zip file for each day, named YYYY-MM-DD.zip
        start_date: the first date that is read from the folder of zip files, inclusive.
            should be in the format YYYY-MM-DD
        start_date: the last date that is read from the folder of zip files, inclusive.
            should be in the format YYYY-MM-DD
        deployment_metadata_filepath: the filepath to the deployment metadata .ini file that will be read
    
    returns:
        tuple of dataframes that can be stored to the database using DBDriver.store()
            one dataframe per table in the database
    """
    
    if deployment_metadata_filepath is None:
        raise ValueError("a deployment_metadata_filepath must be passed")
    
    hsr_dates  = hsr_func.Get_hsr_Dates(hsr_path, start_date, end_date)
    if len(hsr_dates) == 0:
        raise ValueError(f"No data could be found for the supplied dates: {start_date}, {end_date}")
    
    gps_type = "gps"
    gps_file_name = "GPS.txt"
    
    day_one = zipfile.ZipFile(os.path.join(hsr_path, hsr_dates[0]+".zip"))
    if "AccessoryData.txt" in day_one.namelist():
        gps_type = "accessory"
        gps_file_name = "AccessoryData.txt"
    
    
    
    m_Ed = []
    m_Eds = []
    m_Summary = []
    m_Gps = []
    m_IndCh = []
    
    
    for hsr_date in hsr_dates:
        print("reading "+hsr_date+".zip")
        dfTemp = ImportHSRFiles.open_hsr_file(hsr_path, hsr_date, 'Total.txt')
        if len(dfTemp): 
            m_Ed.append( dfTemp)
        dfTemp =  ImportHSRFiles.open_hsr_file(hsr_path, hsr_date, 'Diffuse.txt')
        if len(dfTemp): 
            m_Eds.append(dfTemp)
        dfTemp = ImportHSRFiles.open_hsr_file(hsr_path, hsr_date, 'Summary.txt')
        if len(dfTemp): 
            m_Summary.append(dfTemp)
        dfTemp = ImportHSRFiles.open_hsr_file(hsr_path, hsr_date, 'IndCh.txt')
        if len(dfTemp):
            m_IndCh.append(dfTemp)
        dfTemp = ImportHSRFiles.open_hsr_file(hsr_path, hsr_date, gps_file_name)
        if len(dfTemp):
            m_Gps.append(dfTemp)
    
    
    ed = pd.concat(m_Ed)
    eds = pd.concat(m_Eds)
    summary = pd.concat(m_Summary)  
    ind_ch = pd.concat(m_IndCh)
    
    gps = None
    if len(m_Gps):
        gps = pd.concat(m_Gps)
    
    reformatter = reformatData.ReformatData()
    
    reformatted = reformatter.reformat_data([ed, eds, summary, ind_ch, gps], deployment_metadata_filepath, gps_type)
    
    return reformatted
    


def read_raw_txt(hsr_path="", start_date="2000-01-01", end_date="2100-01-01", deployment_metadata_filepath=None):
    """reads each channel of spectral data from the .txt files
    params:
        hsr_path: the filepath to the data that will be read. 
            this folder should contain one .zip file for each day, named YYYY-MM-DD.zip
        start_date: the first date that is read from the folder of zip files, inclusive.
            should be in the format YYYY-MM-DD
        start_date: the last date that is read from the folder of zip files, inclusive.
            should be in the format YYYY-MM-DD
        deployment_metadata_filepath: the filepath to the deployment metadata .ini file that will be read
    
    returns a tuple of dataframes that can be used with DBDriver.store_raw to make a raw dataset
    """
    if deployment_metadata_filepath is None:
        raise ValueError("a deployment_metadata_filepath must be passed")
    
    hsr_dates  = hsr_func.Get_hsr_Dates(hsr_path, start_date, end_date)
    if len(hsr_dates) == 0:
        print("warning, no dates selected to read")
    
    filenames = []
    for hsr_date in hsr_dates:
        zip_obj = zipfile.ZipFile(os.path.join(hsr_path, hsr_date)+".zip")
        filenames = []
        for name in zip_obj.namelist():
            if "Raw" in name and name not in filenames:
                filenames.append(name)
    
    dfs = [[] for i in range(len(filenames))]
    
    for hsr_date in hsr_dates:
        
        for i, filename in enumerate(filenames):
            df_temp = ImportHSRFiles.open_hsr_file(hsr_path, hsr_date, filename, Raw=True)
            if len(df_temp): 
                dfs[i].append(df_temp)
    
    for i, df in enumerate(dfs):
        dfs[i] = pd.concat(df)
    
    reformatter = reformatData.ReformatData()
    deployment_metadata = reformatter.reformat_deployment_metadata(deployment_metadata_filepath)
    
    return dfs, deployment_metadata


def read_raw_pixels(file_path=None, skiprows=2):
    """reads the raw pixel data for each channel in the .txt files
    params:
        file_path: the filepath to the data that will be read. 
            this folder should contain one .zip file for each day, named YYYY-MM-DD.zip
        skiprows: the nuber of rows at the top of each file to skip when reading
    """
    single = False
    files = None
    if file_path is None:
        files = os.listdir()
    else:
        if os.path.isdir(file_path):
            files = os.listdir(file_path)
        else:
            files = [file_path]
            single = True
    
    all_columns = []
    
    dictlist = []
    for file_name in files:
        file = file_name
        if not single:
            file = file_path + "/" + file_name
        temp_df = pd.read_csv(file, sep="\t", skiprows=skiprows)
        tempdict = {}
        tempdict["pc_time_end_measurement"] = None
        file_columns = list(temp_df.columns)
        
        del file_columns[0]
        
        for col in file_columns:
            if not col in all_columns:
                all_columns.append(col)
        
        for column in file_columns:
            tempdict[column] = temp_df[column].to_numpy()[:-1]
        dictlist.append(tempdict)
    
    df_columns = []
    for file_col in all_columns:
        df_columns.append(file_col.replace("Ch", "channel_"))
    
    
    result = pd.DataFrame(dictlist)
    result.columns=["pc_time_end_measurement"]+df_columns
    return result


def store_segmented(db_driver, hsr_path, start_date, end_date, period=1, precalculated_values=True, temp_databases_location="temp_databases/"):
    """stores a dataset from txt in seperate databases, and merges them"""
    store_seperate(hsr_path, start_date, end_date, period, precalculated_values, temp_databases_location, _raise=True)
    db_driver.db_store.combine_database_folder(temp_databases_location, delete=True)

def store_seperate(hsr_path, start_date, end_date, deployment_metadata_file_path, period=1, precalculated_values=True, temp_databases_location="temp_databases/", db_type="sqlite", _raise=False):
    """stores a dataset into multiple databases"""
    hsr_dates = hsr_func.Get_hsr_Dates(hsr_path, start_date, end_date)
    
    if not os.path.isdir(temp_databases_location):
        os.mkdir(temp_databases_location)
    
    prev_files = os.listdir(temp_databases_location)
    if len(prev_files) > 0:
        if _raise:
            raise FileExistsError("there are already files in the target directory")
        print("WARNING: there are already files in the target directory")
        
    
    list_of_lists = []
    for i in range(0, len(hsr_dates)-period, period):
        list_of_lists.append(hsr_dates[i:i+period])
    list_of_lists.append(hsr_dates[i+period:])
    
    reformatter = reformatData.ReformatData()
    
    for small_hsr_dates in list_of_lists:
        db_name = "temp_db_"+small_hsr_dates[0]+".db"
        db_type = db_type

        temp_driver = DBDriver(temp_databases_location+db_name, db_type, deployment_metadata_file_path)
        
        dfs, gps_type = read(hsr_path, small_hsr_dates[0], small_hsr_dates[-1])
        reformatted = reformatter.reformat_data(dfs, deployment_metadata_file_path, gps_type)
        temp_driver.db_store.store(reformatted, temp_driver)
        if precalculated_values:
            temp_driver.add_precalculated_values()





def store_to_hdf(file_path, ed, eds, summary, hdr, ind_ch, gps):
    ##### makes the folder file_path if it dosent exist
    if not os.path.isdir(file_path):
        print("hdfs folder dosent exist, making a new one")
        strings = file_path.split("/")
        string_path = ""
        for string in strings[:-1]:
            string_path += string+"/"
            if not os.path.isdir(string_path):
                os.mkdir(string_path)
    
    ##### stores dataframes to hdf
    ed.to_hdf(file_path+"ed", "ed")
    eds.to_hdf(file_path+"eds", "eds")
    summary.to_hdf(file_path+"summary", "summary")
    hdr.to_hdf(file_path+"hdr", "hdr")
    ind_ch.to_hdf(file_path+"indch", "indch")
    gps.to_hdf(file_path+"gps", "gps")


def load_from_hdf(file_path):
    ed = pd.read_hdf(file_path+"ed", "ed")
    eds = pd.read_hdf(file_path+"eds", "eds")
    summary = pd.read_hdf(file_path+"summary", "summary")
    hdr = pd.read_hdf(file_path+"hdr", "hdr")
    ind_ch = pd.read_hdf(file_path+"indch", "indch")
    gps = pd.read_hdf(file_path+"gps", "gps")
    
    return [ed, eds, summary, hdr, ind_ch, gps]


# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""
import uuid
import os
import importlib_resources

import numpy as np
import pandas as pd

pvlib_imported = False

try:
    import pvlib
    pvlib_imported = True
except:
    print("pvlib was not successfully imported")

import hsr1
from hsr1.utils.spectrum import spectrumUtils

SCALING_FACTOR = 0.617

rand = np.random.default_rng()

class SyntheticDataset():
    def __init__(self):
        if not pvlib_imported:
            raise ImportError("pvlib is required to generate synthetic datasets")
    
    def __init__(self, spectral_func=None, 
                 start_date="2024-01-01 00:00:00+00:00", 
                 end_date="2024-01-02 00:00:00+00:00",
                 latitude=53.14,
                 longitude=-1.64,
                 altitude=40):
        
        
        if spectral_func is None:
            self.spectral_func = self.generate_spectral_data
        
        self.spectral_data = None
        self.system_data = None
        self.deployment_metadata = None
        self.accessory_data = None
        
        self.start_date = start_date
        self.end_date = end_date
        self.latitude = latitude
        self.longitude = longitude
        self.altitude = altitude
        
        self.dataseries_id = None
    
    def get_dfs(self):
        return [self.spectral_data, self.system_data, self.deployment_metadata, self.accessory_data]
    
    def generate_custom_column(self, table, column, function, *args, **kwargs):
        if table == "spectral_data":
            self.spectral_data[column] = function(self.spectral_data["pc_time_end_measurement"], 
                                                  *args, **kwargs)
        elif table == "system_data":
            self.system_data[column] = function(self.spectral_data["pc_time_end_measurement"], 
                                                *args, **kwargs)
        elif table == "accessory_data":
            self.accessory_data[column] = function(self.accessory_data["pc_time_end_measurement"], 
                                                   *args, **kwargs)
        
        elif table == "deployment_metadata":
            self.deployment_metadata[column] = function(*args, **kwargs)
            
        else:
            raise ValueError("table not found")
    
    def remove_rows(self, table, function, *args, **kwargs):
        if table == "spectral_data":
            self.spectral_data = self.spectral_data.loc[function(self.spectral_data["pc_time_end_measurement"], *args, **kwargs), :]
        if table == "system_data":
            self.system_data = self.system_data.loc[function(self.system_data["pc_time_end_measurement"], *args, **kwargs), :]
        if table == "accessory_data":
            self.accessory_data = self.accessory_data.loc[function(self.accessory_data["pc_time_end_measurement"], *args, **kwargs), :]

    def global_integral(self, timestamps):
        if self.system_data is None:
            raise ValueError("system_data is required, generate that first")
            
        
        
    
    def generate_spectral_data(self, period="min"):
        spectral_data = pd.DataFrame()
        
        spectral_data["pc_time_end_measurement"] = pd.date_range(self.start_date, self.end_date, freq=period).astype(str)
        num_readings = len(spectral_data)
        spectral_data["sample_id"] = self.generate_sample_ids(spectral_data["pc_time_end_measurement"])
        spectral_data["dataseries_id"] = self.generate_dataseries_id()
        spectral_data["global_spectrum"] = self.smarts_integral_data(spectral_data["pc_time_end_measurement"]).tobytes()
        spectral_data["diffuse_spectrum"] = self.smarts_integral_data(spectral_data["pc_time_end_measurement"]).tobytes()
        spectral_data["global_integral"] = pd.to_datetime(spectral_data["pc_time_end_measurement"]).dt.hour
        spectral_data["diffuse_integral"] = spectral_data["global_integral"]/2
        spectral_data["global_molar"] = spectral_data["global_integral"]
        spectral_data["diffuse_molar"] = spectral_data["diffuse_integral"]
        spectral_data["camera_temp"] = rand.normal(30, 0.1, num_readings)
        
        
        self.spectral_data = spectral_data
    
    def generate_system_data(self, period="min"):
        system_data = pd.DataFrame()
        
        system_data["pc_time_end_measurement"] = pd.date_range(self.start_date, self.end_date, freq=period).astype(str)
        system_data["gps_time"] = pd.to_datetime(system_data["pc_time_end_measurement"]).dt.tz_localize(None).astype(str)
        system_data["gps_status"] = "A"
        system_data["gps_latitude"] = self.latitude
        system_data["gps_longitude"] = self.longitude
        system_data["gps_altitude"] = 100
        system_data["pressure"] = 1000
        system_data["baro_temp"] = 25
        system_data["rh"] = 13
        system_data["rh_temp"] = 0.0
        system_data["dataseries_id"] = self.generate_dataseries_id()
        system_data["sample_id"] = self.generate_sample_ids(system_data["pc_time_end_measurement"])
        
        
        self.system_data = system_data
    
    
    def generate_system_data_accessory(self, accessory_data=None, spectral_timestamps=None, spectral_sample_ids=None):
        if accessory_data is None:
            accessory_data = self.accessory_data
        if spectral_timestamps is None:
            spectral_timestamps = self.spectral_data["pc_time_end_measurement"]
        if spectral_sample_ids is None:
            spectral_sample_ids = self.spectral_data["sample_id"]
        
        system_data = hsr1.utils.ReformatData().reformat_system_data_reformatted_accessory(accessory_data, spectral_timestamps, spectral_sample_ids)
        
        system_data["dataseries_id"] = self.generate_dataseries_id()
        system_data["sample_id"] = self.generate_sample_ids(system_data["pc_time_end_measurement"])
        
        self.system_data = system_data
        
        
    
    def generate_deployment_metadata(self, filepath):
        self.deployment_metadata = hsr1.utils.ReformatData().reformat_deployment_metadata(filepath)
        # self.deployment_metadata["default_elevation"] = str(self.altitude)
        # self.deployment_metadata["default_latitude"] = str(self.latitude)
        # self.deployment_metadata["default_longitude"] = str(self.longitude)
    
    def generate_accessory_data(self, period="min"):
        accessory_data = pd.DataFrame()
        
        accessory_data["pc_time_end_measurement"] = pd.date_range(self.start_date, self.end_date, freq=period).astype(str)
        
        num_readings = len(accessory_data)
        
        accessory_data["Clock_Error"] = 0.0
        accessory_data["Latitude"] = rand.normal(self.latitude, 0.1, num_readings)
        accessory_data["Longitude"] = rand.normal(self.longitude, 0.1, num_readings)
        accessory_data["Altitude"] = rand.normal(100.0, 0.1, num_readings)
        accessory_data["GPSAge"] = 0.0
        accessory_data["NSV"] = rand.normal(10.0, 0.1, num_readings).astype(int)
        accessory_data["GPSSpare"] = rand.normal(5.0, 0.1, num_readings)
        accessory_data["_15Vin"] = rand.normal(23000.0, 0.1, num_readings)
        accessory_data["_3VCAP"] = rand.normal(22000.0, 0.1, num_readings)
        accessory_data["HTRIn"] = rand.normal(50.0, 0.1, num_readings)
        accessory_data["Vcc"] = rand.normal(3300.0, 0.1, num_readings)
        accessory_data["VSpare"] = rand.normal(0.0, 0.1, num_readings)
        accessory_data["I_Tot"] = rand.normal(484.0, 0.1, num_readings)
        accessory_data["I_15VPC"] = rand.normal(200.0, 0.1, num_readings)
        accessory_data["I_15VCAM"] = rand.normal(100.0, 0.1, num_readings)
        accessory_data["I_5VPC"] = rand.normal(4.0, 0.1, num_readings)
        accessory_data["ISpare"] = 0.0
        accessory_data["T_CPU"] = rand.normal(30.0, 0.1, num_readings)
        accessory_data["T_Bezel"] = rand.normal(29.0, 0.1, num_readings)
        accessory_data["T_RH"] = 0.0
        accessory_data["T_Baro"] = rand.normal(27.0, 0.1, num_readings)
        accessory_data["RH"] = rand.normal(12.0, 0.1, num_readings)
        accessory_data["Pressure"] = rand.normal(900.0, 0.1, num_readings)
        accessory_data["IMU_Output_Type"] = 2.0
        accessory_data["Yaw"] = rand.normal(358.0, 0.1, num_readings)
        accessory_data["Roll"] = rand.normal(-1.62, 0.1, num_readings)
        accessory_data["Pitch"] = rand.normal(10.31, 0.1, num_readings)
        accessory_data["Q_W"] = 16314.0
        accessory_data["Q_X"] = -1476.0
        accessory_data["Q_Y"] = 219.0
        accessory_data["Q_Z"] = 217.0
        accessory_data["Acc_X"] = 0.02
        accessory_data["Acc_Y"] = -0.23
        accessory_data["Acc_Z"] = 0.06
        accessory_data["IMU1"] = 0.0
        accessory_data["IMU2"] = 0.0
        accessory_data["Control_flags"] = 130.0
        accessory_data["Control_flags2"] = 0.0
        accessory_data["StatusFlags"] = 6144.0
        accessory_data["Heater"] = 0.0
        accessory_data["MsgCount"] = 69.0
        accessory_data["RdgCount"] = 0.0
        accessory_data["Watchdog"] = 3590.0
        accessory_data["gps_time"] = pd.to_datetime(accessory_data["pc_time_end_measurement"]).dt.tz_localize(None).astype(str)
        accessory_data["cpu_time"] = pd.to_datetime(accessory_data["pc_time_end_measurement"]).dt.tz_localize(None).astype(str)
        accessory_data["dataseries_id"] = self.generate_dataseries_id()
        accessory_data["sample_id"] =  self.generate_sample_ids(accessory_data["pc_time_end_measurement"])
        
        self.accessory_data = accessory_data
    
    
    """
    ---------------------------------------------------------------------------
    individual column functions
    ---------------------------------------------------------------------------
    """
    
    
    def generate_sample_ids(self, timeseries):
        """must have spectral_data generated first"""
        new_sample_ids = pd.DataFrame()
        new_sample_ids["pc_time_end_measurement"] = timeseries
        
        
        if self.spectral_data is not None:
            new_sample_ids = new_sample_ids.merge(self.spectral_data[["pc_time_end_measurement", "sample_id"]], how="left", on="pc_time_end_measurement")
        
        if "sample_id" in new_sample_ids.columns:
            not_matched = pd.isnull(new_sample_ids["sample_id"])
            new_sample_ids.loc[not_matched, "sample_id"] = [str(uuid.uuid1()) for i in range(np.sum(not_matched))]
        else:
            new_sample_ids["sample_id"] = [str(uuid.uuid1()) for i in range(len(timeseries))]
        
        return new_sample_ids["sample_id"]
    
    def generate_dataseries_id(self):
        if self.dataseries_id is None:
            self.dataseries_id = str(uuid.uuid1())
        
        return self.dataseries_id
    
    def linear_integral(self, timeseries):
        return np.ones(len(timeseries))

    def dist_from_midday(self, timeseries):
        timeseries = pd.to_datetime(timeseries)
        mins = timeseries.dt.hour*60 + timeseries.dt.minute
        return 720-np.abs(mins-720)

    def pvlib_integral_static(self, timeseries, *args):
        location = pvlib.location.Location(self.latitude, 
                                           self.longitude, 
                                           altitude=self.altitude)
        integrals = location.get_clearsky(pd.DatetimeIndex(timeseries))
        data = integrals[["ghi", "dhi"]].values*SCALING_FACTOR
        
        return data
    
    def smarts_integral_data(self, timeseries, filepath=None):
        if filepath is None:
            res = importlib_resources.files("hsr1.data").joinpath("smarts.txt")
            file = importlib_resources.as_file(res)
            with file as f:
                filepath = f
        
        smarts_data = spectrumUtils.read_simple_file(filepath)
        
        return smarts_data*len(timeseries)
    
    def integral_hour_of_day(self, timeseries):
        timeseries = pd.to_datetime(timeseries)
        return np.array([timeseries.dt.hour*100, timeseries.dt.hour*50]).T
    
    def remove_n_hour(self, timeseries, hour):
        timestamps = pd.to_datetime(timeseries)
        return timestamps.dt.hour != hour
    
    def num_readings(self, timeseries, base=0, scale=1):
        data = [i for i in range(len(timeseries))]
        data = base + np.array(data)*scale
        return data
        
        
"""
GUIDE TO WRITING CUSTOM FUNCTIONS:

dataseries and timeseries is automatically passed in, so the function must expect it
return a list, np array or series of data for the specific column

use it with

dataseries.generate_custom_column(table, column, function, **kwargs)
"""





if __name__ == "__main__":
    db_name = "databases/synthetic.db"
    
    if os.path.exists(db_name):
        os.remove(db_name)
    
    latitude = 13.19
    longitude = -59.54
    altitude = 40
    
    synthetic_dataset = SyntheticDataset(start_date="2023-01-01 00:00:00+00:00", 
                                         end_date="2023-01-03 23:59:00+00:00",
                                         latitude=latitude, longitude=longitude,
                                         altitude=40)
    
    synthetic_dataset.generate_spectral_data("10s")
    synthetic_dataset.generate_system_data()
    synthetic_dataset.generate_deployment_metadata("C:/Users/albie/work/Datasets/Winster 2023/HSR1-002 Winster 2024 Deployment.ini")
    synthetic_dataset.generate_accessory_data()
    
    synthetic_dataset.generate_custom_column("spectral_data", 
                                             ["global_integral", "diffuse_integral"], 
                                             synthetic_dataset.pvlib_integral_static)
    
    dfs = synthetic_dataset.get_dfs()
    
    db = hsr1.DBDriver(db_name)
    db.store(dfs)
    graph = hsr1.Graph(db)
    graph.plot_integral()
    graph.plot_accessory()
    graph.plot_gps()
    graph.plot_dips_summary()


# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""

import matplotlib
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
import pandas as pd
import numpy as np
import math

from hsr1.plots import (graphUtils,
                        flagData)

class ClearnessDensityGraph:
    def __init__(self, cmap=plt.get_cmap("jet"), diffuse_name="dif"):
        self.diffuse_name = diffuse_name
        self.cmap = cmap
        ##### according to brsn limits, highest possible global value is 1407
        self.max_integral = 1407
        self.max_density = 0
        self.match_density = False
        self.requirements = {"ghi_toa":["toa_hi", "global_integral", "sed"],
                             "diff_toa":["toa_hi", "diffuse_integral", "sed"],
                             "dni_toa":["toa_hi", "direct_normal_integral", "sed"],
                             "sza_diff_ghi":["diffuse_integral", "global_integral", "sza"],
                             "ghi_dni_clearness":["global_integral", "direct_normal_integral", "toa_hi", "sza"],
                             "ghi_diff_clearness":["global_integral", "diffuse_integral", "toa_hi"],
                             "all":["toa_hi", "global_integral", "diffuse_integral", "direct_normal_integral", "sed", "sza"]}
        self.limits = flagData.calculate_limits()
        self.flag_all = False
        self.flags = pd.DataFrame()
        self.flags[["ghi_possible_min", "ghi_possible_max", "dif_possible_min", "dif_possible_max", "dni_possible_min", "dni_possible_max"]] = False
    
    
    def clearness_density_graph(self, 
                                axes, 
                                df:pd.DataFrame, 
                                x_axis:str, 
                                y_axis:str, 
                                bins=None, 
                                bin_scale:int=500, 
                                axes_range:[[float]]=None, 
                                max_density:bool=None, 
                                match_density:bool=False,
                                flags:pd.DataFrame=None):
        """plots 2 columns against each other with density shown by colour
        params:
            axes: the axes that will be plotted on
            df: dataframe containing data to be plotted
            x_axis: name of the column containing the x_axis data
            y_axis: name of the column containing the y_axis data
            bins: bins for the histogram, if none, generated using bin_scale https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.hist2d.html
            bin_scale: if bins = None, generate bins using this value
            axes_range: the leftmost and rightmost edges of the bins in each dimension, if none, calculated automatically
            max_density: maximum density for the colormap, used when matching density between plots
            match_density: wether or not to match the density to the max_density variable
            flags: the flags for this specific graph
        """
        cmap = self.cmap
        
        bbox = axes.get_window_extent()
        vert_pixels = int(bbox.height)
        horiz_pixels = int(bbox.width)
        
        if bins is None:
            bins = [min(horiz_pixels, 100), min(vert_pixels, 100)]
        
        
        x_axes_data = df[x_axis].replace({np.inf:np.nan})#, np.NINF:np.nan})
        y_axes_data = df[y_axis].replace({np.inf:np.nan})#, np.NINF:np.nan})
        if axes_range is None:
            axes_range = [[np.nanmin(x_axes_data), np.nanmax(x_axes_data)], [np.nanmin(y_axes_data), np.nanmax(y_axes_data)]]
        
        hist, *_ = np.histogram2d(x_axes_data, y_axes_data, bins=bins, range=axes_range)
        
        if np.max(hist) == 0:
            print("no data to plot on hist: ", x_axis, ", ", y_axis)
            return
        
        self.max_density = max(self.max_density, np.max(hist))
        if match_density:
            if max_density == None:
                max_density = self.max_density
            else:
                self.max_density = max(self.max_density, max_density)
            max_fraction = np.max(hist)/max_density
            cmap = matplotlib.colors.LinearSegmentedColormap.from_list("new_colormap", cmap(np.linspace(0, max_fraction, 100)))
        
        extent = (axes_range[0][0], axes_range[0][1], axes_range[1][0], axes_range[1][1])
        im = axes.imshow(np.flip(hist.T, 0), aspect="auto", interpolation="none", cmap=cmap, extent=extent, norm=matplotlib.colors.LogNorm(vmin=1))
        
        xticks = axes.get_xticks()
        xlabels = axes.get_xticklabels()
        
        
        
        
        if self.flag_all:
            flag_text = str(round((sum(flags.any(axis=1))/len(df))*100, 1)) + "%"
            axes.text(0,0.88*axes.set_ylim()[1], flag_text)
            flags = self.flags
        
        if flags is not None:
            colordict = {
                "red":((0,1,1), (1,1,1)),
                "green":((0,1,1), (1,0,0)),
                "blue":((0,1,1), (1,1,1))}
            flag_cmap = LinearSegmentedColormap("LinearCmap", colordict)
            flagged_data = df.loc[flags.any(axis="columns")]
            if len(flagged_data) > 0:
                flagged_hist, *_ = np.histogram2d(flagged_data[x_axis], flagged_data[y_axis], bins=bins, range=axes_range)
                axes.imshow(np.flip(flagged_hist.T, 0), aspect="auto", interpolation="none", norm=matplotlib.colors.LogNorm(vmin=0, vmax=1), cmap=flag_cmap, extent=extent)
        
        axes.set_xticks(xticks, xlabels)
        
        return im
                
    
    def ghi_toa(self, axes, df:pd.DataFrame, match_density=None):
        """plot to compare the ghi to the top of atmosphere irradiance
        params:
            axes: the axes to plot on
            df: the data to be plotted
            match_density: wether or not to match the density to self.max_integral
        """
        ##### the flags checked by this graph
        flag_names = ["ghi_possible_min", "ghi_possible_max", "ghi_rare_min", "ghi_rare_max"]
        
        if match_density == None:
            match_density = self.match_density
        axes.plot(self.limits["toa_hi"], self.limits["ghi_possible"], color="black")
        axes.plot(self.limits["toa_hi"], self.limits["ghi_rare"], color="black")
        
        flags = None
        if not self.flags.empty:
            flags = self.flags[flag_names]
        
        im = self.clearness_density_graph(axes, df, "toa_hi", "global_integral", 
                                          axes_range=[[0,np.nanmax(df["toa_hi"])],[0,self.max_integral]], 
                                          match_density=match_density, 
                                          flags=flags)
        
        xlim = axes.get_xlim()
        axes.set_xlim(xlim[0], xlim[1]*1.05)
        
        axes.set_ylabel("GHI(W/m²)")
        axes.set_xlabel("Top of Atmosphere(W/m²)")
        
        return im
    
    def diff_toa(self, axes, df:pd.DataFrame, match_density=None):
        """plot to compare the diffuse to the top of atmosphere irradiance
        params:
            axes: the axes to plot on
            df: the data to be plotted
            match_density: wether or not to match the density to self.max_integral
        """
        ##### the flags checked by this graph
        flag_names = ["dif_possible_min", "dif_possible_max", "dif_rare_min", "dif_rare_max"]
        
        if match_density == None:
            match_density = self.match_density
        axes.plot(self.limits["toa_hi"], self.limits["diff_possible"], color="black")
        axes.plot(self.limits["toa_hi"], self.limits["diff_rare"], color="black")
        
        flags = None
        if not self.flags.empty:
            flags = self.flags[flag_names]
        
        im = self.clearness_density_graph(axes, df, "toa_hi", "diffuse_integral", 
                                          axes_range=[[0,np.nanmax(df["toa_hi"])],[0,self.max_integral]], 
                                          match_density=match_density, 
                                          flags=flags)
        
        xlim = axes.get_xlim()
        axes.set_xlim(xlim[0], xlim[1]*1.05)
        
        axes.set_ylabel(self.diffuse_name+"(W/m²)")
        axes.set_xlabel("Top of Atmosphere(W/m²)")
        
        return im
    
    def dni_toa(self, axes, df:pd.DataFrame, match_density=None):
        """plot to compare the dni to the top of atmosphere irradiance
        params:
            axes: the axes to plot on
            df: the data to be plotted
            match_density: wether or not to match the density to self.max_integral
        """
        ##### the flags checked by this graph
        flag_names = ["dni_possible_min", "dni_possible_max", "dni_rare_min", "dni_rare_max"]
        
        if match_density == None:
            match_density = self.match_density
        axes.plot(self.limits["toa_hi"], self.limits["dni_possible"], color="black")
        axes.plot(self.limits["toa_hi"], self.limits["dni_rare"], color="black")
        
        flags = None
        if not self.flags.empty:
            flags = self.flags[flag_names]
        
        im = self.clearness_density_graph(axes, df, "toa_hi", "direct_normal_integral", 
                                          axes_range=[[0,np.nanmax(df["toa_hi"])],[0,self.max_integral]], 
                                          match_density=match_density, 
                                          flags=flags)
        
        xlim = axes.get_xlim()
        axes.set_xlim(xlim[0], xlim[1]*1.05)
        
        axes.set_ylabel("DNI(W/m²)")
        axes.set_xlabel("Top of Atmosphere(W/m²)")
    
        return im
    
    
    def sza_diff_ghi(self, axes, df:pd.DataFrame, match_density=None):
        """plots the diffuse ratio against the solar zenith angle
        params:
            axes: the axes to plot on
            df: the data to be plotted
            match_density: wether or not to match the density to self.max_integral
        """
        ##### the flags checked by this graph
        flag_names = ["diffuse_ratio_sza_possible_max"]
        
        if match_density == None:
            match_density = self.match_density
        # TODO: plot uncheckable values(ghi<50)? would mean no need for cropping, but also miss out on some data
        
        ##### copy df for local scope
        new_df = df.copy()
        
        new_df["diff/ghi"] = new_df["diffuse_integral"]/new_df["global_integral"]
        new_df["deg_sza"] = np.degrees(new_df["sza"])
        
        axes_range = [[0, 90], [0, 1.4]]
        
        axes.plot(np.degrees(self.limits["SZA"]), self.limits["diff/ghi_sza_comparison"], color="black")
        
        flags = None
        if not self.flags.empty:
            flags = self.flags[flag_names]
        
        im = self.clearness_density_graph(axes, new_df, "deg_sza", "diff/ghi", 
                                          axes_range=axes_range, 
                                          match_density=match_density, 
                                          flags=flags)
        
        
        axes.set_ylabel(self.diffuse_name+"/GHI")
        axes.set_xlabel("Solar zenith angle(°)")
        
        return im
    
    def ghi_dni_clearness(self, axes, df:pd.DataFrame, match_density=None):
        """plots the dni/toani ratio against the ghi/toa ratio
        params:
            axes: the axes to plot on
            df: the data to be plotted
            match_density: wether or not to match the density to self.max_integral
        """
        ##### the flags checked by this graph
        flag_names = ["dni_toani_ghi_toa_possible_max"]
        
        if match_density == None:
            match_density = self.match_density
        
        ##### copy df for local scope
        new_df = df.copy()
        
        new_df["dni/toani"] = df["direct_normal_integral"]/(df["toa_hi"]/np.cos(df["sza"]))
        new_df["ghi/toa"] = df["global_integral"]/df["toa_hi"]
        
        axes.plot(self.limits["ghi_toa"], self.limits["dni_toani"], color="black")
        axes.vlines(1.35, 0, 0.95, colors="black")
        
        axes_range = [[0, 1.5], [0,1]]
        
        flags = None
        if not self.flags.empty:
            flags = self.flags[flag_names]
        
        im = self.clearness_density_graph(axes, new_df, "ghi/toa", "dni/toani", 
                                          axes_range=axes_range, 
                                          match_density=match_density, 
                                          flags=flags)
        
        
        axes.set_ylabel("DNI/TOANI")
        axes.set_xlabel("GHI/TOA")
        
        return im
    
    def ghi_diff_clearness(self, axes, df:pd.DataFrame, match_density=None):
        """plots the diffuse ratio against the ghi/toa ratio
        params:
            axes: the axes to plot on
            df: the data to be plotted
            match_density: wether or not to match the density to self.max_integral
        """
        ##### the flags checked by this graph
        flag_names = ["diffuse_ratio_ghi_toa"]
        
        if match_density == None:
            match_density = self.match_density
        
        ##### copy df for local scope
        new_df = df.copy()
        
        new_df["diff/ghi"] = df["diffuse_integral"]/df["global_integral"]
        new_df["ghi/toa"] = df["global_integral"]/df["toa_hi"]
        
        axes.plot(self.limits["ghi/toa"], self.limits["diff/ghi"], color="black")
        
        flags = None
        if not self.flags.empty:
            flags = self.flags[flag_names]
        
        axes_range = [[0, 1.5], [0, 1.4]]
        im = self.clearness_density_graph(axes, new_df, "ghi/toa", "diff/ghi", 
                                     axes_range=axes_range, 
                                     match_density=match_density, 
                                     flags=flags)
        
        
        axes.set_ylabel(self.diffuse_name+"/GHI")
        axes.set_xlabel("GHI/TOA")
        
        return im
        
    
    def graph_all(self, axes, df:pd.DataFrame, fig, match_density:bool=False, flags:pd.DataFrame=None, flag_all:bool=False):
        """plots all the possible graphs of this type
        params:
            axes: list of axes that the graphs will be plotted on
            df: the data to be plotted
            fig: the figure all the graphs are plotted on, for plotting a shared colorbar
            match_density: wether or not to match the density to self.max_integral
            flags: all the flags generated for the dataset
            flag_all: wether to plot all the flagged data on each graph, of just the flags from that graph
        """
        if flag_all:
            self.flag_all = True
        
        # if flags is None:
        self.flags = flags
        
        im = self.ghi_toa(axes[0], df[self.requirements["ghi_toa"]])
        self.diff_toa(axes[1], df[self.requirements["diff_toa"]])
        self.dni_toa(axes[2], df[self.requirements["dni_toa"]])
        self.sza_diff_ghi(axes[3], df[self.requirements["sza_diff_ghi"]])
        self.ghi_dni_clearness(axes[4], df[self.requirements["ghi_dni_clearness"]])
        self.ghi_diff_clearness(axes[5], df[self.requirements["ghi_diff_clearness"]])
        
        if match_density:
            fig.colorbar(im, ax=axes, aspect=25, pad=0.01)
        


# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""

from configparser import ConfigParser
import os

class Config:
    def __init__(self, filepath:str):
        self.filepath = filepath
    
    def read_section(self, section_name:str) -> dict:
        """reads a section of the initialised config file
        params:
            section_name: the name of the section to be read
        returns:
            config: a dictionary of all the fields in the relevant section
        """
        if not os.path.isfile(self.filepath):
            raise Exception(f"config file \"{self.filepath}\" not found")
        
        parser = ConfigParser()
        parser.read(self.filepath)
        
        config = {}
        
            
        if parser.has_section(section_name):
            params = parser.items(section_name)
            
            for param in params:
                config[param[0]] = param[1]

        else:
            print("section not found in config file")
            return
        
        return config


# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""
import datetime

import numpy as np
import pandas as pd
import matplotlib

import hsr1.utils.HSRFunc as HsrFunc
from hsr1.plots import graphUtils


class SpectrumGraph:
    def __init__(self, cmap=None, timedelta=pd.Timedelta(0), spectrometer_average_period=60):
        self.cmap = cmap
        self.timedelta = timedelta
        self.spectrometer_average_period = spectrometer_average_period
        self.requirements = {"all":["pc_time_end_measurement", "global_spectrum", "diffuse_spectrum", "sza", "azimuth", "toa_hi"],
                             "ghi":["pc_time_end_measurement", "global_spectrum", "sza", "azimuth", "toa_hi"],
                             "diff":["pc_time_end_measurement", "diffuse_spectrum", "sza", "azimuth", "toa_hi"],
                             "dni":["pc_time_end_measurement", "global_spectrum", "diffuse_spectrum", "sza", "azimuth", "toa_hi"]}
        self.first_time = pd.to_datetime("00:00:00")
        self.last_time = pd.to_datetime("23:59:59")
        self.fig = None
    
    def spectrum_graph(self, df:pd.DataFrame, data_col:str, axes, max_reading, cbar=True, output_location=None):
        if len(df) == 0:
            return
        
        df = df.copy()
        df["pc_time_end_measurement"] = df["pc_time_end_measurement"].dt.tz_localize(None)
        first_date_reading = df["pc_time_end_measurement"].iloc[0].date()
        last_date_reading = df["pc_time_end_measurement"].iloc[len(df)-1].date()
        df = df[np.logical_and(df["pc_time_end_measurement"]>=datetime.datetime.combine(first_date_reading, self.first_time), 
                               df["pc_time_end_measurement"]<=datetime.datetime.combine(last_date_reading, self.last_time))]
        
        if len(df) == 0:
            return
    
        bbox = axes.get_window_extent()
        vert_pixels = int(bbox.height)
        horiz_pixels = int(bbox.width)
        
        
        
        pc_time_secs = (df["pc_time_end_measurement"].dt.tz_localize(None).astype("int64")/10**9).values.astype(int)
        
        first_timestamp = df["pc_time_end_measurement"].iloc[0]
        last_timestamp = df["pc_time_end_measurement"].iloc[-1]
        first_date = first_timestamp.date()
        last_date = last_timestamp.date()
        left_xlim = graphUtils.total_seconds(datetime.datetime.combine(first_date, self.first_time))
        right_xlim = graphUtils.total_seconds(datetime.datetime.combine(last_date, self.last_time))
        
        periods_in_dataset = ((right_xlim-left_xlim)//self.spectrometer_average_period)+1
        num_bins = int(min(periods_in_dataset, horiz_pixels))
        
        pixels = np.linspace(left_xlim, right_xlim, num_bins).astype(int)
        
        full_df = pd.DataFrame()
        full_df["time"] = pixels
        full_df["data"] = [np.full(801, np.nan)]*len(pixels)
        groups = np.clip(np.digitize(pc_time_secs, pixels), 0, len(pixels)-1)
        full_df.loc[groups, "data"] = df[data_col].values
        
        data = full_df["data"].values
        data = np.stack(data).T
        
        
        
        extent = (left_xlim, 
                  right_xlim, 
                  1100, 300)
        
        image = axes.imshow(data, cmap=self.cmap, norm=None, vmin=0, vmax=max_reading, aspect="auto",
                            interpolation="nearest", extent=extent)
        
        axes.set_ylabel(data_col)
        
        axes.set_xlim(left_xlim, right_xlim)
        
        if first_date == last_date:
            xticks, xlabels = self.generate_hour_xticks(first_date, last_date)
            axes.set_xticks(xticks, xlabels)
        else:
            xticks = axes.get_xticks()
            xticks = xticks[np.logical_and(xticks>=left_xlim, xticks<=right_xlim)]
            xlabels = []
            for i in range(len(xticks)):
                xlabels.append(datetime.datetime.fromtimestamp(xticks[i]).time())
            axes.set_xticks(xticks, xlabels)
        
        if cbar:
            self.add_colourmap(axes, max_reading, max(1, max_reading//5))
    
    def supplementary_integral_plot(self, data, axes, first_time, last_time, day, spec_day, cmap=False):
        ##### plot line graph of global and diffuse integral
        data = data[["pc_time_end_measurement", "global_integral", "diffuse_integral"]].copy()
        data["pc_time_end_measurement"] = data["pc_time_end_measurement"].dt.tz_localize(None)
        first_date_reading = data["pc_time_end_measurement"].iloc[0].date()
        last_date_reading = data["pc_time_end_measurement"].iloc[len(data)-1].date()
        left_xlim = datetime.datetime.combine(first_date_reading, first_time)
        right_xlim = datetime.datetime.combine(last_date_reading, last_time)
        data = data[np.logical_and(data["pc_time_end_measurement"]>=left_xlim, 
                                   data["pc_time_end_measurement"]<=right_xlim)]
        
        pc_time_secs = (data["pc_time_end_measurement"].astype("int64")/10**9).values.astype(int)
        axes[3].plot(pc_time_secs, data["global_integral"], label="global integral")
        axes[3].plot(pc_time_secs, data["diffuse_integral"], label="diffuse integral")
        axes[3].legend()
        xticks, xlabels = spec_day.generate_hour_xticks(day, day)
        axes[3].set_xticks(xticks, xlabels)
        
        axes[3].set_xlim(graphUtils.total_seconds(left_xlim), graphUtils.total_seconds(right_xlim))
        
        if cmap:
            spec_day.add_colourmap(axes[3], 0, invisible=True)
    
    def add_colourmap(self, axes, max_reading, gap=0.5, invisible=False):
        shrink = 1 if not invisible else 0.0001
        cbar = self.fig.colorbar(matplotlib.cm.ScalarMappable(cmap="jet"), ax=axes, aspect=25, pad=0.01, shrink=shrink)
        yticks = np.arange(0, max_reading, gap)
        ytick_location = yticks/max_reading
        ytick_location[ytick_location == np.inf] = 0
        cbar.ax.set_yticks(ytick_location, yticks)
    
    def generate_hour_xticks(self, first_date, last_date):
        first_timestamp_hour = datetime.datetime.combine(first_date, datetime.time(hour=self.first_time.hour+1))
        last_timestamp_hour = datetime.datetime.combine(last_date, datetime.time(hour=self.last_time.hour))
        hours = pd.date_range(first_timestamp_hour, last_timestamp_hour, freq="1h")
        xticks = []
        xlabels = []
        for hour in hours:
            xticks.append(hour.timestamp())
            xlabels.append(hour.hour)
        xticks = np.array(xticks)
        xlabels = np.array(xlabels)
        return xticks, xlabels
    
    
    def direct_normal_spectrum(self, data:pd.DataFrame, axes, max_reading):
        self.spectrum_graph(data, "direct_normal_spectrum", 
                            axes, max_reading, True)
    
    def global_spectrum(self, data, axes, max_reading):
        self.spectrum_graph(data, "global_spectrum", 
                            axes, max_reading, True)
    
    def diffuse_spectrum(self, data, axes, max_reading):
        self.spectrum_graph(data, "diffuse_spectrum", 
                            axes, max_reading, True)
    
    
    def plot_all(self, data, axes_list):
        self.global_spectrum(data, axes_list[0], 2)
        self.diffuse_spectrum(data, axes_list[1], 10)
        self.direct_normal_spectrum(data, axes_list[2], 2)
        
        
    
    
    def plot_all_aod(self, data, axes_list, max_reading):
        self.spectrum_graph(data, "total_od", axes_list[0], max_reading, False)
        self.spectrum_graph(data, "aod_microtops", axes_list[1], max_reading, False)
        self.spectrum_graph(data, "aod_wood_2017", axes_list[2], max_reading, False)
        
        cbar = self.fig.colorbar(matplotlib.cm.ScalarMappable(cmap="jet"), ax=axes_list, aspect=25, pad=0.01)
        yticks = np.arange(0, max_reading, 0.5)
        ytick_location = yticks/max_reading
        ytick_location[ytick_location == np.inf] = 0
        cbar.ax.set_yticks(ytick_location, yticks)


# -*- coding: utf-8 -*-
"""
copyright 2024 Peak Design
this file is part of hsr1, which is distributed under the GNU Lesser General Public License v3 (LGPL)
"""

import numpy as np
import pandas as pd
import scipy as sp
import matplotlib.pyplot as plt
import lmfit

from hsr1.utils.spectrum.pixelSpectrum import PixelSpectrum
from hsr1.utils.spectrum.weightedSpectrumSegment import WeightedSpectrumSegment

import hsr1.utils.spectrum.spectrumUtils as SpectrumUtils


class WeightedSpectrum:
    def __init__(self, weighted_spectrum=None, wavelength_spectra=None, 
                 weighted_pixel_spectrum=None, pixel_spectra=None, 
                 weighted_reference=None, reference_spectrum=None, 
                 applied_poly=None, amplitude_calibration=None, 
                 default_poly=[1623, -1.4, 0, 0],
                 filter_low_values=False, num_reference_dips=30, 
                 pixel_spectrum=None, 
                 name="", find_peaks=False,
                 graph_limits=[300, 1101],
                 **kwargs):
        """represents the location and prominence of dips over the whole dataset
        params:
            weighted_spectrum: if you alredy have a weighted spectrum, you can just pass it in here
            wavelength_spectra: dataframe to be converted to a weighted spectrum
            weighted_pixel_spectrum: weighted spectrum in pixel space. 
                        will have a default calibration applied to it to convert into wavelength space, to compare with the reference
            pixel_spectra: dataframe containing pixel data, will be converted into a wavelength spectrum and then a weighted spectrum
            weighted_reference: weighted reference spectrum
            reference_spectrum: spectrum to be converted to weighted_reference
            lin_cal: for manually finding the first dip, input how much to move the measured spectrum by
                        (not usually needed but maybe helpful for very uncalibrated spectra)
            applied_poly: the polynomial that has already been applied to this spectrum
            filter_low_values: wether or not to filter out spectra with low total intensities,
                        as bright days will have clearer spectrums.
            pixel_spectrum: an example pixel spectrum, ideally taken from this dataset
                        not used for calculation, but helpful for plotting the calibration
            name: name to display on graphs
        """
        self.reference = reference_spectrum
        self.weighted_reference = weighted_reference
        self.wavelength_spectra = wavelength_spectra
        self.pixel_spectra = pixel_spectra
        
        if np.max(pixel_spectrum) == 0:
            print("channel " + name + " is all zeros")
        else:
            self.pixel_spectrum = pixel_spectrum/np.max(pixel_spectrum)
        
        self.applied_poly = applied_poly
        self.amplitude_calibration = amplitude_calibration
        
        self.ps = PixelSpectrum(0)
        
        self.default_poly = default_poly
        
        self.filter_low_values = filter_low_values
        self.find_peaks = find_peaks
        
        self.num_reference_dips = num_reference_dips
        
        self.spectrum, self.source = self.calculate_spectrum(weighted_spectrum, 
                                                             wavelength_spectra,
                                                             weighted_pixel_spectrum,
                                                             pixel_spectra)
        
        
        if self.amplitude_calibration is not None:
            self.spectrum = self.ps.apply_amplitude_calibration(self.spectrum, self.amplitude_calibration)
        
        if self.weighted_reference is None:
            self.weighted_reference = pd.DataFrame({"global_spectrum":[self.reference]})
            self.weighted_reference = self.generate_weighted_frequency(self.weighted_reference, n=num_reference_dips)
            
        self.reference_peaks = self.find_n_peaks(self.weighted_reference)
        
        
        
        self.calibration = None
        self.calibrated_spectrum = None
        
        self.poly_cal = None
        
        self.name = name
        self.graph_limits = graph_limits
    
    
    def calculate_spectrum(self, weighted_spectrum=None, wavelength_spectra=None, weighted_pixel_spectrum=None, pixel_spectra=None):
        """calculates the weighted spectrum given all the possible input types"""
        if weighted_spectrum is not None:
            return weighted_spectrum, "weighted_wavelength_spectrum"
        
        if weighted_pixel_spectrum is not None:
            return weighted_pixel_spectrum, "weighted_pixel_spectrum"
        
        if wavelength_spectra is not None:
            calc_weighted_spectrum = self.generate_weighted_frequency(wavelength_spectra, filter_low_values=self.filter_low_values)
            return calc_weighted_spectrum, "wavelength spectra"
        
        if pixel_spectra is not None:
            self.applied_poly = self.default_poly
            wavelength_spectra = self.ps.apply_wavelength_calibration(pixel_spectra, self.applied_poly, inverse=True)
            calc_weighted_spectrum = self.generate_weighted_frequency(wavelength_spectra, filter_low_values=self.filter_low_values) 
            return calc_weighted_spectrum, "pixel_spectra"
        
    
    def generate_weighted_frequency(self, global_spectrum=None, column_name="global_spectrum", cutoff_angle=np.radians(80), cutoff_wavelength=1000, n=0, filter_low_values=False):
        """generates a list of all wavelengths with the sum of their prominences every time they are one of the top
        
        params:
            global_spectrum: if not loading from database, the dataset to use.
                                Should be a dataframe with columns "pc_time_end_measurement" and "global_spectrum"
            column_name: if loading from database, the column containing the spectral data
            cutoff_angle: if loading from database, the zenith angle to include(radians)
            cutoff_wavelength: the meax wavlength to include, usefull for cropping out extreme noise
            n: how many dips to take from each measurement, if 0, uses all
                                if too low when working with a single spectrum, not enough dips and some segments will be empty(errors)
            filter_low_values: wether or not to filter out spectra with low total intensities, 
                                used to filter out night time data, but will also filter out shaded data
        returns:
            weighted_frequency: a graph showing where dips appear over the dataset, weighted by their prominence
        """
        ##### load from databse if no spectrum is provided
        if global_spectrum is None or len(global_spectrum) == 0:
            ##### some databses may not have precalculated data loaded
            try:
                global_spectrum = self.driver.db_load.load(["pc_time_end_measurement", column_name], condition="sza < "+str(cutoff_angle))
            except:
                global_spectrum = self.driver.db_load.load(["pc_time_end_measurement", column_name])
                
            global_spectrum = global_spectrum.rename(columns={column_name:"global_spectrum"})
            global_spectrum = self.driver.reformat.deserialise(global_spectrum, "global_spectrum")
        
        if filter_low_values:
            selection = []
            for j in range(len(global_spectrum)):
                if np.sum(global_spectrum.loc[j, "global_spectrum"]) > 10000:
                    selection.append(True)
                else:
                    selection.append(False)
            global_spectrum = global_spectrum.loc[selection, :].reset_index()
        
        global_spectrum = global_spectrum["global_spectrum"].values
        
        nlargest, nlargest_prominence = [], []
        
        if n==0:
            ##### if no restrictions, find all peaks
            nn = (None, None)
            for spectrum in global_spectrum:
                if not self.find_peaks:
                    spectrum = -spectrum
                largest, properties = sp.signal.find_peaks(spectrum, prominence=nn)
                nlargest += list(largest)
                nlargest_prominence += list(properties["prominences"])
            nlargest = np.array(nlargest)
            nlargest_prominence = np.array(nlargest_prominence)
            
        else:
            nlargest, nlargest_prominence = SpectrumUtils.find_nlargest(global_spectrum, n, ["wavelengths", "prominences"], cutoff_wavelength, offset=0, find_peaks=self.find_peaks)
        
        weighted_frequency = np.zeros(801)
        ##### add the found dips to the weighted spectrum
        for i in range(len(nlargest)):
            weighted_frequency[nlargest[i]] += nlargest_prominence[i]
        
        
        return weighted_frequency
    
    
    def calculate_segment_calibration(self, centre_point, dist_from_centre, 
                                      c0_lim=20, c1_lim=0.05, iterations=30, 
                                      init_c0_lim=50, init_c0=0, 
                                      plot_attempts=False):
        """loops through the spectrum and finds the linear calibration for each segment
        params:
            centre_point: The point in the reference spectrum where the calibration starts from. Ideally should be a large, identifiable dip.
            dist_from_centre: When splitting the spectrum into segments, how far the segments should extend from the centre of the segment
                            controls the segment width
            c0_lim: When finding a segment's linear calibration, the range of values to check while shifting left/right
            c1_lim: When finding a segment's linear calibration, the range of stretches to test.
            iterations: How many different stretch values to try within the limit (every integer shift is tried within the limits)
            initial_c0_limit: how far to look from the reference centre_point to look for the corresponding peak
            plot_attempts: plot the individual calibrated segment and all the attempted calibration values with their fit quality
        """
        c0 = 0
        c1 = 1
        
        if init_c0 is not None:
            c0 = init_c0
        
        result = {}
        attempts = {}
        attempts_params = {}
        
        if centre_point < 0 or centre_point > 800:
            raise ValueError("centre_point must be within spectrum")
        
        ##### create centre segment and calculate calibration
        segment = WeightedSpectrumSegment(self.weighted_reference, self.spectrum, 
                                          centre_point, dist_from_centre, centre_point)
        first_cal, new_attempts = segment.find_best_calibration(c0, c1, init_c0_lim, c1_lim, iterations=iterations, plot=False, debug=True)
        
        ##### adds the result and metadata to a list
        result[centre_point] = first_cal
        attempts[centre_point] = new_attempts
        attempts_params[centre_point] = [0, 1]
        
        
        ##### right side
        ##### initial values are from centre segment
        right_centre_point = centre_point
        new_cal = [first_cal[0], first_cal[1]]
        while right_centre_point < len(self.spectrum)-2*dist_from_centre:
            old_centre_point = right_centre_point
            right_centre_point += 2*dist_from_centre
            
            attempts_params[right_centre_point] = new_cal
            
            ##### create a new segment to the right of the previous segment
            ##### use the gradient from previous segment as a starting point
            segment = WeightedSpectrumSegment(self.weighted_reference, self.spectrum, right_centre_point, dist_from_centre, old_centre_point)
            

            new_cal, new_attempts = segment.find_best_calibration(new_cal[0], new_cal[1], c0_lim, c1_lim, iterations=iterations, debug=True)
            
            
            result[right_centre_point] = new_cal
            attempts[right_centre_point] = new_attempts
        
        
        
        ##### left side
        ##### initial values are from centre segment
        left_centre_point = centre_point
        new_cal = [first_cal[0], first_cal[1]]
        while left_centre_point > 2*dist_from_centre:
            old_centre_point = left_centre_point
            left_centre_point -= 2*dist_from_centre
            
            attempts_params[left_centre_point] = new_cal
            
            ##### create a new segment to the left of the previous segment
            ##### use the gradient from previous segment as a starting point
            segment = WeightedSpectrumSegment(self.weighted_reference, self.spectrum, left_centre_point, dist_from_centre, old_centre_point)
            
            new_cal, new_attempts = segment.find_best_calibration(new_cal[0], new_cal[1], c0_lim, c1_lim, iterations=iterations, debug=True)
            
            result[left_centre_point] = new_cal
            attempts[left_centre_point] = new_attempts

        self.calibration = result
        
        
        
        if plot_attempts:
            self.plot_attempts(attempts, attempts_params, c0_lim, c1_lim, iterations, dist_from_centre, init_c0_lim)
        

    
    def combine_segments(self, dist_from_centre):
        """combines the calibrated segments into one spectrum
        params:
            dist_from_centre: distance from the centre to the edge of the segment
        """
        ##### creates a spectrum of zeros that will be overwritten
        final_spectrum = np.zeros(len(self.spectrum))
        
        ##### loops through each centre point
        keys = list(self.calibration.keys())
        keys.sort()
        for key in keys:
            ##### calibrates the individual segment
            new_cal = self.calibration[key]
            calibrated = self.ps.apply_origin_wavelength_calibration(self.spectrum, (0, new_cal[1], 0, 0), key)
            calibrated = self.ps.apply_constant_wavelength_calibration(calibrated, new_cal[0])
            
            #### adds the calibrated segment to the final sepctrum
            left_edge = max(0, key-dist_from_centre)
            right_edge = min(len(calibrated), key+dist_from_centre)
            final_spectrum[left_edge:right_edge] = calibrated[left_edge:right_edge]
        
        self.calibrated_spectrum = final_spectrum
    
    
    
    def find_n_peaks(self, spectrum, n=None):
        """find the n largest peaks in a spectrum
        params:
            spectrum: spectrum containing peaks
            n: number of peaks to find. if none returns all
        returns:
            reference_peaks: the index of all the peaks
        """
        spectrum = spectrum/max(np.abs(spectrum))
        peaks, properties = sp.signal.find_peaks(spectrum, prominence=(None, None))
        prominences = properties["prominences"]
        
        ##### sorts by most prominent peak
        index = np.flip(prominences.argsort())
        
        if n == None:
            n = len(peaks)-1
        n = min(len(peaks)-1, n)

        reference_peaks = peaks[index][:n]
        return reference_peaks
    
    def decalibrate_linear_segment(self, ref_peak, cal_peak):
        """undoes the linear calibration at a point"""
        cal = self.calibration
        centres = np.array(list(cal.keys()))
        
        ##### finds which calibration has been applied
        closest_centre_index = np.argmin(np.absolute(centres-ref_peak))
        closest_centre = centres[closest_centre_index]
        segment_cal = cal[closest_centre]
        
        ##### calibration was done around the centre point, so find position relative to centre_point
        #####   then decalibrate around 0, then add back the centre point
        centre_dist = cal_peak-closest_centre
        decalibrated = -SpectrumUtils.apply_inverse_polynomial(centre_dist, segment_cal[0], segment_cal[1], segment_cal[2], segment_cal[3])
        
        decalibrated_peak = closest_centre - decalibrated
        
        return decalibrated_peak
    
    
    def match_fit_quality(self, params, ref, measured):        
        c0 = params["c0"].value
        c1 = params["c1"].value
        c2 = params["c2"].value
        c3 = params["c3"].value
        return ref-SpectrumUtils.apply_inverse_polynomial(measured, c0, c1, c2, c3)
    
    def match_dips(self, reference_data, measured_data, method="least_squares"):
        params = lmfit.Parameters()
        params.add('c0', value=1620)#, min=-20, max=20)
        params.add('c1', value=-1.4)#, min=-1.6, max=-1.2)
        params.add('c2', value=0.0)#, min=-0.0000001, max=0.00000001)
        params.add('c3', value=0.0)
        ##### limit c3  to be negative
        ##### when applying inverse polynomials, 
        #####   there is no way to be sure which value to pick when interpolating from a non 1-1 function(one that has seperate maxima and minima)
        ##### so the polynomials are limited to decreasing functions, which will always be 1-1 in this case
        
        result = lmfit.minimize(self.match_fit_quality, params, kws = {"ref":reference_data, "measured":measured_data}, method=method)
        
        ##### unpack results
        c0_result = result.params["c0"].value
        c1_result = result.params["c1"].value
        c2_result = result.params["c2"].value
        c3_result = result.params["c3"].value
        
        return [c0_result, c1_result, c2_result, c3_result]
    
    def match_n_dips(self, n, left_limit=0, nlargest=None, reference_dip_locations=None):
        """matches dips between the reference and the linear calibrated spectrum, then converts to pixel position
        params:
            n: number of dips to match
            left_limit: the leftmost dip to consider, as SMARTS spectrum extends further left than is measured
            nlargest: how many peaks to choose from from the calibrated spectrum, can help avoid matching with tiny dips
        returns:
            reference_peaks: the matched_dips' location in the reference spectrum
            measured_peaks: the corresponding position in the uncalibrated pixel spectrum
        """
        ##### finds all the largest peaks from the reference data
        calibrated_peaks = self.find_n_peaks(self.calibrated_spectrum, nlargest)
        
        reference_peaks = np.array(reference_dip_locations)
        if reference_dip_locations is None:
            ##### takes the n largest peaks from the reference data that are past the left_limit
            reference_peaks = self.reference_peaks[self.reference_peaks >= left_limit][:n]
        else:
            n = len(reference_dip_locations)
            reference_peaks -= 300
        
        calibrated_peaks = np.sort(calibrated_peaks)
        reference_peaks = np.sort(reference_peaks)
        
        ##### finds the closest measured peaks to the reference peaks
        matched_peaks_index = [np.argmin(np.absolute(calibrated_peaks-peak)) for peak in reference_peaks]
        matched_peaks = calibrated_peaks[matched_peaks_index]
        
        # print(matched_peaks)
        
        ##### convert the found peaks back into pixel data
        measured_peaks= []
        for i in range(len(reference_peaks)):
            ref_peak = reference_peaks[i]
            cal_peak = matched_peaks[i]
            
            ##### undoes the linear calibration
            decalibrated_peak = self.decalibrate_linear_segment(ref_peak, cal_peak)
            # print(decalibrated_peak)
            
            ref_peak += 300
            decalibrated_peak += 300
            
            ##### undoes the previous calibration
            if self.applied_poly is not None:
                decalibrated_peak = SpectrumUtils.apply_polynomial(decalibrated_peak, self.applied_poly[0], self.applied_poly[1], self.applied_poly[2], self.applied_poly[3])
            
            measured_peaks.append(decalibrated_peak)
        measured_peaks = np.array(measured_peaks)
        
        # print(measured_peaks)
        
        return reference_peaks, measured_peaks
    
    
    
    def plot_attempts(self, attempts, attempts_params, c0_lim, c1_lim, iterations, dist_from_centre, init_c0_lim):
        """For each segment, plots the found linear calibration and a histogram of all tested value pairs
        params:
            attempts: a dictionary with keys corresponding to the centre of each segment,
                        containing a 2d array of the fit quality for each tested value
            attempt_params: a dictionary with keys corresponding to the centre of each segment,
                        containing the starting point for each segment's calibration [c0, c1]
            c0_lim: When finding a segment's linear calibration, the range of values to check while shifting left/right
            c1_lim: When finding a segment's linear calibration, the range of stretches to test.
            dist_from_centre: When splitting the spectrum into segments, how far the segments should extend from the centre of the segment
                        controls the segment width
            iterations: How many different stretch values to try within the limit (every integer shift is tried within the limits)
            init_c0_lim: initial c0 limit when finding main peak
        """
        unsorted_keys = list(self.calibration.keys())
        centre = unsorted_keys[0]
        keys = unsorted_keys.copy()
        keys.sort()
        for key in keys:
            new_cal = self.calibration[key]
            calibrated = self.ps.apply_origin_wavelength_calibration(self.spectrum, (0, new_cal[1], 0, 0), key)
            calibrated = self.ps.apply_constant_wavelength_calibration(calibrated, new_cal[0])
            
            ##### calculate the windowed reference to plot with the calibrated spectrum
            reference = self.weighted_reference/max(self.weighted_reference)
            segment = reference[max(0, key-dist_from_centre):key+dist_from_centre]
            
            left_padding_dist = max(0, key-dist_from_centre)
            right_padding_dist = max(0, len(self.weighted_reference)-(key+dist_from_centre))
            
            windowed_reference = np.concatenate((np.zeros(left_padding_dist), segment, np.zeros(right_padding_dist)))
            if max(windowed_reference) > 0:
                windowed_reference = windowed_reference/max(windowed_reference)
            
            ##### plot the spectrum after the found calibration and the weighted reference
            plt.figure(figsize=(20, 10))
            plt.plot(windowed_reference, label="reference")
            plt.plot(calibrated/max(calibrated), label="calibrated")
            plt.axvline(key, c="black")
            plt.axvline(key+dist_from_centre, c="gray")
            plt.axvline(key-dist_from_centre, c="gray")
            plt.suptitle("Shift: "+str(new_cal[0]) + "\n Stretch: " + str(np.round(new_cal[1], 2)))
            plt.legend()
            plt.show()
            
            ##### prepare data to plot all attempted shift and stretch values with fit quality
            this_attempts = attempts[key]
            this_attempts = np.flip(this_attempts, axis=0)
            init_values = attempts_params[key]
            
            plt.figure(figsize=(20,10))
            plt.imshow(this_attempts)
            
            ##### if centre and c0 known, plot hist as line not rectangle
            single = False
            if len(this_attempts[0]) == 1:
                single = True
            
            this_c0_lim = c0_lim
            if key == centre:
                this_c0_lim = init_c0_lim
            
            ##### calculates the xticks and labels and plots them
            if not single:
                shift_range = np.arange(init_values[0]-this_c0_lim, init_values[0]+this_c0_lim+1)
                x = np.linspace(0, len(shift_range)-1, 9).astype(int)
                plt.xticks(x, shift_range[x])
                plt.xlabel("shift")
            if single:
                plt.xticks([])
            
            ##### plots the yticks and labels
            stretch_range = np.linspace(init_values[1]-c1_lim, init_values[1]+c1_lim, iterations).round(3)
            stretch_range = np.flip(stretch_range)
            y = np.linspace(0, len(stretch_range)-1, 9).astype(int)
            plt.yticks(y, stretch_range[y])
            plt.ylabel("stretch")
            plt.suptitle(self.name+"\n segment centre point: "+str(key))
            
            plt.show()
    
    def plot_segmented_calibration(self, dist):
        """plots all the linear calibrations in the segment they are used"""
        segmented_cal = self.calibration
        keys = list(segmented_cal.keys())
        
        plt.figure(figsize=(20,10))
        for key in keys:
            poly = segmented_cal[key]
            x_data = np.arange(key-dist, key+dist)
            y_data = []
            for i in range(len(x_data)):
                y_data.append(SpectrumUtils.apply_polynomial(x_data[i], poly[0], poly[1], poly[2], poly[3]))
            y_data = np.array(y_data).round().astype(int)
            plt.scatter(x_data, y_data, s=1, c="tab:blue")
        plt.suptitle(self.name+" "+ "segmented wavelength calibration")
        plt.show()
    
    def plot_segmented_linear_calibration(self, dist_from_centre):
        """plot the combination of all segments with their linear calibration, 
        a first guess at the calibration so dips can be matched"""
        weighted_reference = np.concatenate((np.zeros(300), self.weighted_reference))
        calibrated_spectrum = np.concatenate((np.zeros(300), self.calibrated_spectrum))
        
        plt.figure(figsize=(20, 10))
        plt.plot(weighted_reference/max(weighted_reference), label="reference")
        # plt.plot(self.spectrum/max(self.spectrum), label="measured")
        plt.plot(calibrated_spectrum/max(calibrated_spectrum), label="segmented linear calibration")
        
        for wavelength in self.calibration:
            wavelength += 300
            plt.axvline(wavelength, color="lightgray")
            plt.axvline(wavelength+dist_from_centre, color="black")
            plt.axvline(wavelength-dist_from_centre, color="black")
            plt.annotate((str(round(self.calibration[wavelength-300][0]))+", "+str(round(self.calibration[wavelength-300][1], 4))+"x"), [wavelength, 1])
        
        plt.xlim(self.graph_limits[0], self.graph_limits[1])
        plt.yticks([])
        plt.ylabel("wavelength")
        plt.suptitle(self.name+" "+ "initial segmented linear calibration")
        plt.legend()
        plt.show()
    
    def plot_matched_dips(self, ref, measured):
        """plots the dips that have been found onto their corrseponding spectra, 
        usedul for checking that the right dips have been matched to each other"""
        ref_spectrum = np.concatenate((np.zeros(300), self.weighted_reference/max(self.weighted_reference)))
        
        ##### use wavelength spectrums if they are available
        if self.reference is not None:
            ref_spectrum = np.concatenate((np.zeros(300), self.reference/max(self.reference)))
        
        ref_string = "\nReference dips:   "
        measured_string = "\nMeasured dips: "
        for _ref, _measured in zip(ref.astype(str), measured.astype(str)):
            ref_string += str(_ref) + ","
            ref_string += " "*(4-len(_ref)) if len(_ref) < 4 else ""
            measured_string += str(_measured) + ","
            measured_string += " "*(4-len(_measured)) if len(_measured) < 4 else ""
        
        
        measured = np.flip(measured)
        
        plt.figure(figsize=(20, 10))
        plt.plot(ref_spectrum, label="reference")
        plt.plot(self.pixel_spectrum, label="measured")
        plt.scatter(ref, ref_spectrum[ref.astype(int)])
        plt.scatter(measured, self.pixel_spectrum[measured])
        plt.ylim(0, 1)
        plt.yticks([])
        plt.ylabel("wavelength")
        plt.suptitle(self.name+" "+ "selected dips" + ref_string[:-2] + measured_string[:-2])
        plt.legend()
        plt.show()
    
    def plot_calibration(self, measured, poly):
        """plots the calibration polynomial, with the dips used to find it"""
        calibrated_dips = np.array([SpectrumUtils.apply_inverse_polynomial(x, poly[0], poly[1], poly[2], poly[3]) for x in measured]).round().astype(int)
        cal_string = f"{poly[3]:.2e}x^3 + {poly[2]:.2e}x^2 + {poly[1]:.2f}x + {poly[0]:.2f}"
        
        plt.figure(figsize=(20, 10))
        plt.scatter(measured, calibrated_dips)
        if self.applied_poly is not None:
            plt.plot(self.ps.apply_wavelength_calibration(np.arange(0, 1200), self.applied_poly, inverse=True), np.arange(300, 1101), label="old calibration")
        plt.plot(self.ps.apply_wavelength_calibration(np.arange(0, 1200), poly, inverse=True), np.arange(300, 1101), label="calibration")
        plt.xlim(0, 1200)
        plt.suptitle(self.name+" "+ "calibration: \n" + cal_string)
        plt.xlabel("pixels")
        plt.ylabel("wavelength")
        plt.legend()

        plt.show()

    
    def get_calibrated_spectrum(self, poly):
        calibrated_spectrum = self.ps.apply_wavelength_calibration(self.pixel_spectrum, poly, inverse=True, flip=True)
        if self.amplitude_calibration is not None:
            calibrated_spectrum = self.ps.apply_amplitude_calibration(calibrated_spectrum, self.amplitude_calibration)
            calibrated_spectrum = calibrated_spectrum/max(calibrated_spectrum)
        calibrated_spectrum = np.concatenate((np.zeros(300), calibrated_spectrum))
        return calibrated_spectrum
    
    def plot_calibrated_spectrum(self, poly, comp_ref, comp_measured, reference_dip_locations):
        """calibrates the pixel spectrum with the old and the new calibration and plots them against the reference"""
        
        plt.figure(figsize=(20,10))
        
        ##### adds 300 zeros to the start of the spectrum and sets the xlim at 300, so the graph starts at 300, not 0
        reference = np.concatenate((np.zeros(300), self.reference))
        plt.plot(reference/max(reference), label="reference")
        
        original_label = "original calibration"
        if self.source == "pixel_spectra":
            original_label = "default calibration"
        original_calibrated_spectrum = self.get_calibrated_spectrum(self.applied_poly)
        plt.plot(original_calibrated_spectrum, label=original_label)
        
        calibrated_spectrum = self.get_calibrated_spectrum(poly)
        plt.plot(calibrated_spectrum/max(calibrated_spectrum), label="calibrated")
        
        for dip in comp_ref:
            plt.axvline(dip, c="lightgray")
        
        
        
        
        plt.xlim(self.graph_limits[0], self.graph_limits[1])
        plt.ylim(0, 1)
        plt.yticks([])
        plt.ylabel("amplitude")
        plt.xlabel("wavelength")
        plt.suptitle(self.name+" "+ "First measurement calibrated")
        plt.legend()
        plt.show()
        
        # calibrated_dataset = self.ps.apply_wavelength_calibration(self.wavelength_spectra, poly, np.arange(300, 1101), output_shape=[300, 1101])
        
        # graph = Graph()
        # graph.daily_peaks_density(calibrated_dataset, reference_lines=np.sort(ref), reference_labels=[])
        
        # try:
        #     graph.daily_peaks_density(calibrated_dataset, reference_lines=np.sort(ref), reference_labels=[])
        # except ValueError as e:
        #     if e == "histogram needs multiple days of data":
        #         print("pass more data to get summary plot")
        #     else:
        #         print("unknown exception")
    
    
    def compare_dips(self, poly, reference_dip_locations, left_limit, n):
        calibrated_spectrum = self.get_calibrated_spectrum(poly)
        
        reference_peaks = np.array(reference_dip_locations)
        if reference_dip_locations is None:
            ##### takes the n largest peaks from the reference data that are past the left_limit
            reference_peaks = self.reference_peaks[self.reference_peaks >= left_limit][:n]
        else:
            reference_peaks -= 300
        reference_peaks += 300
        calibrated_peaks = self.find_n_peaks(-calibrated_spectrum+max(calibrated_spectrum))
        
        
        calibrated_peaks = np.sort(calibrated_peaks)
        reference_peaks = np.sort(reference_peaks)
        
        if len(calibrated_peaks) == 0:
            return np.array([]), np.array([])
        
        
        
        ##### finds the closest measured peaks to the reference peaks
        matched_peaks_index = [np.argmin(np.absolute(calibrated_peaks-peak)) for peak in reference_peaks]
        matched_peaks = calibrated_peaks[matched_peaks_index]
        
        return reference_peaks, matched_peaks
        
        
        
    
    def plot_error(self, poly, ref, measured, unrounded):
        unrounded = [SpectrumUtils.apply_inverse_polynomial(x, poly[0], poly[1], poly[2], poly[3]) for x in unrounded]
        
        if len(ref) != len(unrounded) or len(measured) != len(unrounded):
            return
        
        error = measured-ref
        
        plt.figure(figsize=(20, 10))
        plt.plot(ref, error)
        plt.scatter(ref, error)
        for x, y in zip(ref, error):
            plt.annotate(f"{x:.0f}", (x, y))
        plt.plot(ref, np.sort(unrounded)-ref)
        plt.show()
    
    
    
    def find_calibration(self, centre_point=760, dist_from_centre=50, 
                         num_dips=15, 
                         c0_lim=10, c1_lim=0.1, iterations=50, 
                         init_c0_lim=30, init_c0=0, 
                         left_cutoff=360, 
                         reference_dip_locations=None, 
                         plot_debug=False, plot_results=True, plot_summary=False,
                         return_dips=False, 
                         **kwargs):
        """finds the best calibration from pixels to wavelengths.
        
        params:
            centre_point: The point in the reference spectrum where the calibration starts from. Ideally should be a large, identifiable dip.
            init_c0_limit: if lin_cal is not set, how far to look from the reference centre_point to look for the corresponding peak
            dist_from_centre: When splitting the spectrum into segments, how far the segments should extend from the centre of the segment
                            controls the segment width
            left_cutoff: When selecting dips to match, the minimum reference wavelength to pick, 
                            as the measured spectrum drops off at 350nm,whereas SMARTS has data down to 300nm, 
                            so dips could being selected that arent in the measured data.
                            Should be the reference equivalent of the measured limit
            num_dips: number of dips to match between reference data and measured data
            c0_lim: When finding a segment's linear calibration, the range of values to check while shifting left/right
            c1_lim: When finding a segment's linear calibration, the range of stretches to test.
            iterations: How many different stretch values to try within the limit (every integer shift is tried within the limits)
            plot_debug: For each segment, plots the found linear calibration and a histogram of all tested value pairs
            plot_results: plots a selection of graphs giving an ovewrview of each step of the process and the accuracy of the calibration
        
        returns:
            poly: list containing the values that are the coefficint for the found polynomial
                            [c0, c1, c2, c3]
        """
        if not self.num_reference_dips >= num_dips:
            raise ValueError("num_reference_dips must be >= num_dips")
            
        
        ##### convert from wavelength to index
        centre_point -= 300
        left_cutoff -= 300
        
        ##### calculate the linear calibration for each segment of the spectrum
        self.calculate_segment_calibration(centre_point, dist_from_centre, c0_lim=c0_lim, c1_lim=c1_lim, iterations=iterations, init_c0_lim=init_c0_lim, init_c0=init_c0, plot_attempts=plot_debug)
        ##### combines the linear calibrated segments into one spectrum that is close to being calibrated
        self.combine_segments(dist_from_centre)
        if sum(self.calibrated_spectrum>0) < num_dips:
            if return_dips:
                return None, None
            return None
        ##### find n pairs of dips that correspond between the reference data and the pixel data
        if reference_dip_locations is not None:
            num_dips = len(reference_dip_locations)
        
        ref, measured = self.match_n_dips(num_dips, left_cutoff, reference_dip_locations=reference_dip_locations)
        if len(measured) != num_dips:
            if return_dips:
                return None, None
            return None
        
        unrounded = measured
        measured = measured.round().astype(int)
        
        
        
        ##### converting from index space to wavelength space
        ref += 300
        
        poly = self.match_dips(ref, measured)
        
        self.poly_cal = poly
        
        comp_ref, comp_measured = self.compare_dips(poly, reference_dip_locations, left_cutoff, num_dips)
        
        
        if plot_results:
            self.plot_segmented_linear_calibration(dist_from_centre)
            self.plot_matched_dips(ref, measured)
            self.plot_calibration(measured, poly)
            self.plot_calibrated_spectrum(poly, comp_ref, comp_measured, reference_dip_locations)
            self.plot_error(poly, comp_ref, comp_measured, unrounded)
        elif plot_summary:
            self.plot_calibrated_spectrum(poly, comp_ref, comp_measured, reference_dip_locations)
            
        if return_dips:
            return poly, (comp_ref, comp_measured)
        return poly
        
        
        